# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from typing import Dict, Tuple

from basana.backtesting import config, errors, value_map
from basana.core import helpers as core_helpers
from basana.core.bar import Bar, BarEvent
from basana.core.pair import Pair


class Prices:
    def __init__(self, bid_ask_spread_pct: Decimal, config: config.Config):
        assert bid_ask_spread_pct > Decimal(0)

        self._bid_ask_spread_pct = bid_ask_spread_pct
        self._config = config
        self._last_bars: Dict[Pair, Bar] = {}

    def on_bar_event(self, event: BarEvent):
        self._last_bars[event.bar.pair] = event.bar

    def get_bid_ask(self, pair: Pair) -> Tuple[Decimal, Decimal]:
        last_bar = self._last_bars.get(pair)
        if not last_bar:
            raise errors.NoPrice(f"No price for {pair}")

        last_price = last_bar.close
        pair_info = self._config.get_pair_info(pair)
        half_spread = core_helpers.truncate_decimal(
            (last_price * self._bid_ask_spread_pct / Decimal("100")) / Decimal(2),
            pair_info.quote_precision
        )
        bid = last_price - half_spread
        ask = last_price + half_spread
        return bid, ask

    def get_price(self, pair: Pair) -> Decimal:
        last_bar = self._last_bars.get(pair)
        if not last_bar:
            raise errors.NoPrice(f"No price for {pair}")
        return last_bar.close

    def convert(self, amount: Decimal, from_symbol: str, to_symbol: str) -> Decimal:
        if amount == Decimal(0):
            return Decimal(0)

        for pair, price_fun in [
                (Pair(from_symbol, to_symbol), lambda price: price),
                (Pair(to_symbol, from_symbol), lambda price: Decimal(1) / price),
        ]:
            last_bar = self._last_bars.get(pair)
            if last_bar:
                return amount * price_fun(last_bar.close)
        raise errors.NoPrice(f"No price to convert from {from_symbol} to {to_symbol}")

    def convert_value_map(self, values: value_map.ValueMapDict, to_symbol: str) -> Decimal:
        ret = Decimal(0)
        for symbol, value in values.items():
            if symbol != to_symbol:
                value = self.convert(value, symbol, to_symbol)
            ret += value
        return ret



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union
from urllib.parse import urlparse, urljoin
import asyncio

import aiohttp

from basana.core import token_bucket, helpers as core_helpers
from basana.core.config import get_config_value
from basana.external.bitstamp import config, helpers


class Error(Exception):
    def __init__(self, msg: str, resp: aiohttp.ClientResponse, json_response):
        super().__init__(msg)
        self.status = resp.status
        self.reason = resp.reason
        self.json_response = json_response


def raise_for_error(resp: aiohttp.ClientResponse, json_response):
    msg = None
    if isinstance(json_response, dict):
        if json_response.get("status") == "error":
            msg = json_response.get("reason", "")
        elif error := json_response.get("error"):
            msg = error
        elif (code := json_response.get("code")) and json_response.get("errors"):
            msg = code
    if msg is None and not resp.ok:
        msg = "{} {}".format(resp.status, resp.reason)
    if msg is not None:
        raise Error(msg, resp, json_response)


class APIClient:
    def __init__(
            self, api_key: Optional[str] = None, api_secret: Optional[str] = None,
            session: Optional[aiohttp.ClientSession] = None, tb: Optional[token_bucket.TokenBucketLimiter] = None,
            config_overrides: dict = {}
    ):
        assert not ((api_key is None) ^ (api_secret is None)), \
            "Both api_key and api_secret should be set, or none of them"

        self._api_key = api_key
        self._api_secret = api_secret
        self._session = session
        self._tb = tb
        self._config_overrides = config_overrides

    async def _make_request(
            self, method: str, path: str, authenticate: bool, qs_params: Dict[str, Any] = {},
            data: Dict[str, Any] = {}
    ) -> Any:
        # Throttling enabled ? go sleep if necessary.
        if self._tb and (sleep_time := self._tb.consume()):
            await asyncio.sleep(sleep_time)

        async with core_helpers.use_or_create_session(session=self._session) as session:
            headers = {}
            skip_auto_headers = ["Content-Type"]
            session_method = {
                "GET": session.get,
                "POST": session.post,
            }[method]
            base_url = get_config_value(config.DEFAULTS, "api.http.base_url", overrides=self._config_overrides)
            timeout = get_config_value(config.DEFAULTS, "api.http.timeout", overrides=self._config_overrides)
            url = urljoin(base_url, path)

            if authenticate:
                assert self._api_key is not None and self._api_secret is not None, "Missing credentials"
                hostname = urlparse(base_url).hostname
                assert hostname is not None, "hostname not set"
                nonce = helpers.generate_nonce()
                headers = helpers.get_auth_headers(
                    hostname, self._api_key, self._api_secret, nonce, method, path, qs_params=qs_params, data=data
                )
            async with session_method(
                    url, headers=headers, skip_auto_headers=skip_auto_headers, params=qs_params, data=data,
                    timeout=timeout
            ) as resp:
                json_response = None
                if (ct := resp.headers.get("Content-Type")) and ct.lower().find("application/json") == 0:
                    json_response = await resp.json()
                raise_for_error(resp, json_response)
                return json_response

    async def get_trading_pairs_info(self) -> List[Dict]:
        return await self._make_request("GET", "/api/v2/trading-pairs-info/", False)

    async def get_order_book(self, currency_pair: str, group: Optional[int] = None) -> dict:
        params = {}
        if group is not None:
            params = {"group": group}
        return await self._make_request("GET", f"/api/v2/order_book/{currency_pair}/", False, qs_params=params)

    async def get_ticker(self, currency_pair: str) -> dict:
        return await self._make_request("GET", f"/api/v2/ticker/{currency_pair}/", False)

    async def get_ohlc_data(
            self, currency_pair: str, step: int, limit: int, start: Optional[int] = None, end: Optional[int] = None,
            exclude_current_candle: bool = False
    ) -> dict:
        assert start is None or end is None, "both start and end should not be set"

        params: Dict[str, Any] = {
            "step": step,
            "limit": limit,
        }
        set_optional_params(params, (
            ("start", start),
            ("end", end),
        ))
        if exclude_current_candle:
            params["exclude_current_candle"] = "true"
        return await self._make_request("GET", f"/api/v2/ohlc/{currency_pair}/", False, qs_params=params)

    async def get_websocket_auth_token(self) -> dict:
        return await self._make_request("POST", "/api/v2/websockets_token/", True)

    async def get_account_balances(self) -> List[dict]:
        return await self._make_request("POST", "/api/v2/account_balances/", True)

    async def get_account_balance(self, currency: str) -> dict:
        return await self._make_request("POST", f"/api/v2/account_balances/{currency}/", True)

    async def get_open_orders(self, currency_pair: Optional[str] = None) -> List[dict]:
        url = "/api/v2/open_orders/{}/".format("all" if currency_pair is None else currency_pair)
        return await self._make_request("POST", url, True)

    async def get_order_status(
            self, id: Optional[Union[str, int]] = None, client_order_id: Optional[str] = None,
            omit_transactions: Optional[bool] = None
    ) -> dict:
        assert (id is not None) ^ (client_order_id is not None), "Either id or client_order_id should be set"

        data: Dict[str, Any] = {"id": id} if id else {"client_order_id": client_order_id}
        if omit_transactions:
            data["omit_transactions"] = omit_transactions
        return await self._make_request("POST", "/api/v2/order_status/", True, data=data)

    async def cancel_order(self, id: Union[str, int]) -> dict:
        return await self._make_request("POST", "/api/v2/cancel_order/", True, data={"id": id})

    async def create_market_order(
            self, action: str, currency_pair: str, amount: Decimal, client_order_id: Optional[str] = None,
            **kwargs: Dict[str, Any]
    ) -> dict:
        assert action in ["buy", "sell"], "Invalid action"

        data: Dict[str, Any] = {
            "amount": str(amount),
        }
        if client_order_id:
            data["client_order_id"] = client_order_id
        data.update(kwargs)
        return await self._make_request("POST", f"/api/v2/{action}/market/{currency_pair}/", True, data=data)

    async def create_limit_order(
            self, action: str, currency_pair: str, amount: Decimal, price: Decimal,
            client_order_id: Optional[str] = None, **kwargs: Dict[str, Any]
    ) -> dict:
        assert action in ["buy", "sell"], "Invalid action"

        data: Dict[str, Any] = {
            "amount": str(amount),
            "price": str(price),
        }
        if client_order_id:
            data["client_order_id"] = client_order_id
        data.update(kwargs)
        return await self._make_request("POST", f"/api/v2/{action}/{currency_pair}/", True, data=data)

    async def create_instant_order(
            self, action: str, currency_pair: str, amount: Decimal, amount_in_counter: bool = False,
            client_order_id: Optional[str] = None, **kwargs: Dict[str, Any]
    ) -> dict:
        assert action in ["buy", "sell"], "Invalid action"
        assert amount_in_counter is False or action == "sell", "amount_in_counter only supported for sell orders"

        data: Dict[str, Any] = {
            "amount": str(amount),
        }
        set_optional_params(data, (
            ("client_order_id", client_order_id),
            ("amount_in_counter", amount_in_counter),
        ))
        data.update(kwargs)
        return await self._make_request("POST", f"/api/v2/{action}/instant/{currency_pair}/", True, data=data)


def set_optional_params(params: Dict[str, Any], tuples: Sequence[Tuple[str, Any]]):
    for k, v in tuples:
        if v is None:
            continue
        # if isinstance(v, Decimal):
        #     v = str(v)
        params[k] = v



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from typing import Any, Awaitable, Callable
import datetime
import logging

from . import helpers
from basana.core import event, websockets as core_ws
from basana.core.pair import Pair


logger = logging.getLogger(__name__)


class Trade:
    def __init__(self, pair: Pair, json: dict):
        assert json["e"] == "trade"

        #: The trading pair.
        self.pair: Pair = pair
        #: The JSON representation.
        self.json: dict = json

    @property
    def id(self) -> str:
        """The trade id."""
        return str(self.json["t"])

    @property
    def datetime(self) -> datetime.datetime:
        """The trade datetime."""
        return helpers.timestamp_to_datetime(int(self.json["T"]))

    @property
    def price(self) -> Decimal:
        """The price."""
        return Decimal(self.json["p"])

    @property
    def amount(self) -> Decimal:
        """The amount."""
        return Decimal(self.json["q"])

    @property
    def buy_order_id(self) -> str:
        """The buyer order id."""
        return str(self.json["b"])

    @property
    def sell_order_id(self) -> str:
        """The seller order id."""
        return str(self.json["a"])


class TradeEvent(event.Event):
    """An event for new trades.

    :param when: The datetime when the event occurred. It must have timezone information set.
    :param trade: The trade.
    """

    def __init__(self, when: datetime.datetime, trade: Trade):
        super().__init__(when)
        #: The trade.
        self.trade: Trade = trade


# Generate TradeEvent events from websocket messages.
class WebSocketEventSource(core_ws.ChannelEventSource):
    def __init__(self, pair: Pair, producer: event.Producer):
        super().__init__(producer=producer)
        self._pair: Pair = pair

    async def push_from_message(self, message: dict):
        event = message["data"]
        self.push(TradeEvent(
            helpers.timestamp_to_datetime(int(event["E"])),
            Trade(self._pair, event)
        ))


def get_channel(pair: Pair) -> str:
    return "{}@trade".format(helpers.pair_to_order_book_symbol(pair).lower())


TradeEventHandler = Callable[[TradeEvent], Awaitable[Any]]



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import cast, Any, Awaitable, Callable, Dict, Generator, List, Optional, Set, Tuple
import abc
import asyncio
import contextlib
import dataclasses
import datetime
import heapq
import logging
import platform
import signal

from . import dt, errors, event, helpers, logs


logger = logging.getLogger(__name__)
EventHandler = Callable[[event.Event], Awaitable[Any]]
IdleHandler = Callable[[], Awaitable[Any]]
SchedulerJob = Callable[[], Awaitable[Any]]


@dataclasses.dataclass
class EventDispatch:
    event: event.Event
    handlers: List[EventHandler]


@dataclasses.dataclass(order=True)
class ScheduledJob:
    when: datetime.datetime
    job: SchedulerJob = dataclasses.field(compare=False)  # Comparing function objects fails on Win32


class SchedulerQueue:
    def __init__(self):
        self._queue = []

    def push(self, when: datetime.datetime, job: SchedulerJob):
        assert not dt.is_naive(when), f"{when} should have timezone information set"
        heapq.heappush(self._queue, ScheduledJob(when=when, job=job))

    def peek_next_event_dt(self) -> Optional[datetime.datetime]:
        ret = None
        if self._queue:
            ret = self._queue[0].when
        return ret

    def peek_last_event_dt(self) -> Optional[datetime.datetime]:
        ret = None
        if self._queue:
            ret = self._queue[-1].when
        return ret

    def pop(self) -> Tuple[datetime.datetime, SchedulerJob]:
        assert self._queue
        scheduled_job = heapq.heappop(self._queue)
        return scheduled_job.when, scheduled_job.job


class EventMultiplexer:
    def __init__(self) -> None:
        self._prefetched_events: Dict[event.EventSource, Optional[event.Event]] = {}

    def add(self, source: event.EventSource):
        self._prefetched_events.setdefault(source)

    def peek_next_event_dt(self) -> Optional[datetime.datetime]:
        self._prefetch()

        next_dt = None
        prefetched_events = [evnt for evnt in self._prefetched_events.values() if evnt]
        if prefetched_events:
            next_dt = min(map(lambda evnt: evnt.when, prefetched_events))
        return next_dt

    def pop(self, max_dt: datetime.datetime) -> Tuple[Optional[event.EventSource], Optional[event.Event]]:
        ret_source: Optional[event.EventSource] = None
        ret_event: Optional[event.Event] = None

        # Find the next event to return, this is, the oldest one that is <= max_dt.
        for source, evnt in self._prefetched_events.items():
            # Prefetch the event for sorting purposes.
            if evnt is None:
                evnt = source.pop()
                self._prefetched_events[source] = evnt
            # If the event matches the filter, check if the next one to return.
            if evnt and evnt.when <= max_dt and (ret_event is None or evnt.when < ret_event.when):
                ret_source = source
                ret_event = evnt

        # Consume the event.
        if ret_source:
            self._prefetched_events[ret_source] = None

        return (ret_source, ret_event)

    def pop_while(self, max_dt: datetime.datetime) -> Generator[Tuple[event.EventSource, event.Event], None, None]:
        while None not in (src_and_event := self.pop(max_dt)):
            yield (cast(event.EventSource, src_and_event[0]), cast(event.Event, src_and_event[1]))

    def _prefetch(self):
        sources_to_pop = [
            source for source, event in self._prefetched_events.items() if event is None
        ]
        for source in sources_to_pop:
            if event := source.pop():
                self._prefetched_events[source] = event


class EventDispatcher(metaclass=abc.ABCMeta):
    """Responsible for connecting event sources to event handlers and dispatching events in the right order.

    :param max_concurrent: The maximum number of events to process concurrently.

    .. note::

        The following helper functions are provided to build event dispatchers suitable for backtesting or for live
        trading:

        * :func:`basana.backtesting_dispatcher`
        * :func:`basana.realtime_dispatcher`
    """

    def __init__(self, max_concurrent: int):
        self._event_handlers: Dict[event.EventSource, List[EventHandler]] = {}
        self._sniffers_pre: List[EventHandler] = []
        self._sniffers_post: List[EventHandler] = []
        self._producers: Set[event.Producer] = set()
        self._active_tasks: Optional[helpers.TaskGroup] = None
        self._running = False
        self._stopped = False
        self._scheduler_queue = SchedulerQueue()
        self._event_mux = EventMultiplexer()
        # Used to execute event and scheduler handlers.
        self._handlers_task_pool = helpers.TaskPool(max_concurrent)
        # Set to True for the dispatcher to stop if a handler raises an exception.
        self.stop_on_handler_exceptions = False

    @property
    def current_event_dt(self) -> Optional[datetime.datetime]:  # pragma: no cover
        helpers.deprecation_warning("Use now() instead")
        return self.now()

    @abc.abstractmethod
    def now(self) -> datetime.datetime:
        raise NotImplementedError()

    @property
    def stopped(self) -> bool:
        """Returns True if stop was called, False otherwise."""
        return self._stopped

    def stop(self):
        """Requests the event dispatcher to stop the event processing loop."""
        logger.debug("Stop requested")
        self._stopped = True
        if self._active_tasks:
            self._active_tasks.cancel()
        self._handlers_task_pool.cancel()

    def subscribe(self, source: event.EventSource, event_handler: EventHandler):
        """Registers an async callable that will be called when an event source has new events.

        :param source: An event source.
        :param event_handler: An async callable that receives an event.
        """

        assert not self._running, "Subscribing once we're running is not currently supported."

        self._event_mux.add(source)
        handlers = self._event_handlers.setdefault(source, [])
        if event_handler not in handlers:
            handlers.append(event_handler)
        if source.producer:
            self._producers.add(source.producer)

    def subscribe_all(self, event_handler: EventHandler, front_run: bool = False):
        """Registers an async callable that will be called for all events.

        :param event_handler: An async callable that receives an event.
        :param front_run: True to front run all other handlers, False otherwise.
        """

        assert not self._running, "Subscribing once we're running is not currently supported."

        sniffers = self._sniffers_pre if front_run else self._sniffers_post
        if event_handler not in sniffers:
            sniffers.append(event_handler)

    def schedule(self, when: datetime.datetime, job: SchedulerJob):
        """Schedules a function to be executed at a given time.

        :param when: The datetime when the function should be execution.
        :param job: The function to execute.
        """
        self._scheduler_queue.push(when, job)

    async def run(self, stop_signals: List[int] = [signal.SIGINT, signal.SIGTERM]):
        """Executes the event dispatch loop.

        :param stop_signals: The signals that will be handled to request :func:`run()` to :func:`stop()`.

        This method will execute the following steps in order:

        #. Call :meth:`basana.Producer.initialize` on all producers.
        #. Call :meth:`basana.Producer.main` on all producers and execute event dispatch loop until stopped.
        #. Call :meth:`basana.Producer.finalize` on all producers.
        """

        assert not self._running, "Can't run twice."
        assert self._active_tasks is None

        # This block has coverage on all platforms except on Windows.
        if platform.system() != "Windows":  # pragma: no cover
            for stop_signal in stop_signals:
                asyncio.get_event_loop().add_signal_handler(stop_signal, self.stop)

        self._running = True
        try:
            # Initialize producers.
            async with self._task_group() as tg:
                for producer in self._producers:
                    tg.create_task(producer.initialize())
            # Run producers and dispatch loop.
            async with self._task_group() as tg:
                for producer in self._producers:
                    tg.create_task(producer.main())
                tg.create_task(self._dispatch_loop())
        except asyncio.CancelledError:
            if not self.stopped:
                raise
        finally:
            # Cancel any pending task in the event pool.
            self._handlers_task_pool.cancel()
            await self._handlers_task_pool.wait()
            # No more cancelation at this point.
            self._active_tasks = None
            # Finalize producers.
            await gather_no_raise(*[producer.finalize() for producer in self._producers])

    def on_error(self, error: Any):
        logger.error(error)

    @abc.abstractmethod
    async def _dispatch_loop(self):
        raise NotImplementedError()

    @contextlib.asynccontextmanager
    async def _task_group(self):
        try:
            async with helpers.TaskGroup() as tg:
                self._active_tasks = tg  # So it can be canceled.
                yield tg
        finally:
            self._active_tasks = None

    async def _dispatch_event(self, event_dispatch: EventDispatch):
        logger.debug(logs.StructuredMessage(
            "Dispatching event", when=event_dispatch.event.when, type=helpers.classpath(event_dispatch.event)
        ))
        if self._sniffers_pre:
            await asyncio.gather(
                *[self._call_event_handler(event_dispatch.event, handler) for handler in self._sniffers_pre]
            )
        if event_dispatch.handlers:
            await asyncio.gather(
                *[self._call_event_handler(event_dispatch.event, handler) for handler in event_dispatch.handlers]
            )
        if self._sniffers_post:
            await asyncio.gather(
                *[self._call_event_handler(event_dispatch.event, handler) for handler in self._sniffers_post]
            )

    async def _call_event_handler(self, event: event.Event, handler: EventHandler):
        try:
            return await handler(event)
        except Exception as e:
            logger.exception(logs.StructuredMessage(
                "Unhandled exception in event handler", error=e, event=dict(type=type(event), when=event.when),
                handler=handler
            ))
            if self.stop_on_handler_exceptions:
                self.stop()

    async def _execute_scheduled(self, dt: datetime.datetime, job: SchedulerJob):
        logger.debug(logs.StructuredMessage("Executing scheduled job", scheduled=dt))

        try:
            await job()
        except Exception as e:
            logger.exception(logs.StructuredMessage(
                "Unhandled exception in handler", error=e, dt=dt, scheduler_job=job
            ))
            if self.stop_on_handler_exceptions:
                self.stop()


class BacktestingDispatcher(EventDispatcher):
    """Event dispatcher for backtesting.

    :param max_concurrent: The maximum number of events to process concurrently.
    """

    def __init__(self, max_concurrent: int):
        super().__init__(max_concurrent=max_concurrent)
        self._last_dt: Optional[datetime.datetime] = None

    async def run(self, stop_signals: List[int] = [signal.SIGINT, signal.SIGTERM]):
        with logs.backtesting_log_mode(self):
            await super().run(stop_signals=stop_signals)

    @property
    def now_available(self) -> bool:
        return self._last_dt is not None

    def now(self) -> datetime.datetime:
        if self._last_dt is None:
            raise errors.Error("Can't calculate current datetime since no events were processed")
        return self._last_dt

    def _set_now(self, now: datetime.datetime):
        # For testing purposes.
        assert self._last_dt is None or now >= self._last_dt
        self._last_dt = now

    async def _dispatch_loop(self):
        while not self.stopped:
            next_dt = self._event_mux.peek_next_event_dt()
            if next_dt:
                # Check that events are processed in ascending order.
                assert self._last_dt is None or next_dt >= self._last_dt, \
                    f"{next_dt} can't be dispatched after {self._last_dt}"

                await self._dispatch_scheduled(next_dt)
                await self._dispatch_events(next_dt)
            else:
                # Dispatch all pending scheduled before stopping.
                if last_scheduled_dt := self._scheduler_queue.peek_last_event_dt():
                    await self._dispatch_scheduled(last_scheduled_dt)
                self.stop()

    async def _dispatch_scheduled(self, dt: datetime.datetime):
        # Execute jobs that were scheduled to run before dt.
        next_scheduled_dt = self._scheduler_queue.peek_next_event_dt()
        while next_scheduled_dt and next_scheduled_dt <= dt:
            # If self._last_dt is already set in the future, don't move it backwards in time.
            next_scheduled_dt, job = self._scheduler_queue.pop()
            if self._last_dt is None or next_scheduled_dt > self._last_dt:
                self._last_dt = next_scheduled_dt

            await self._handlers_task_pool.push(self._execute_scheduled(next_scheduled_dt, job))
            # Waiting here and not outside of the loop to prevent executing distant scheduled jobs at the same time.
            await self._handlers_task_pool.wait()

            next_scheduled_dt = self._scheduler_queue.peek_next_event_dt()

    async def _dispatch_events(self, dt: datetime.datetime):
        # Pop events, push them into the task pool, and wait those to finish executing.
        self._last_dt = dt
        for source, evnt in self._event_mux.pop_while(dt):
            await self._handlers_task_pool.push(
                self._dispatch_event(EventDispatch(event=evnt, handlers=self._event_handlers.get(source, [])))
            )
        await self._handlers_task_pool.wait()


class RealtimeDispatcher(EventDispatcher):
    """Event dispatcher for live trading.

    :param max_concurrent: The maximum number of events to process concurrently.
    """

    def __init__(self, max_concurrent: int):
        super().__init__(max_concurrent=max_concurrent)
        self._prev_event_dt: Dict[event.EventSource, datetime.datetime] = {}
        self.idle_sleep = 0.01
        self._wait_all_timeout: Optional[float] = 0.01
        self._idle_handlers: List[IdleHandler] = []

    def now(self) -> datetime.datetime:
        return dt.utc_now()

    def subscribe_idle(self, idle_handler: IdleHandler):
        """Registers an async callable that will be called when there are no events to dispatch.

        :param idle_handler: An async callable that receives no arguments.
        """

        assert not self._running, "Subscribing once we're running is not currently supported."

        if idle_handler not in self._idle_handlers:
            self._idle_handlers.append(idle_handler)

    async def _dispatch_loop(self):
        while not self.stopped:
            now = dt.utc_now()
            # Feed the task pool with scheduled jobs and events that are ready for processing.
            await asyncio.gather(
                self._push_scheduled(now),
                self._push_events(now),
            )
            # Give some time for tasks to execute, and keep on pushing tasks.
            await self._handlers_task_pool.wait(timeout=self._wait_all_timeout)
            if self._handlers_task_pool.idle:
                await self._on_idle()

    async def _on_idle(self):
        if self._idle_handlers:
            await gather_no_raise(*[
                self._handlers_task_pool.push(idle_handler()) for idle_handler in self._idle_handlers
            ])
        else:
            # Otherwise we'll monopolize the event loop.
            await asyncio.sleep(self.idle_sleep)

    async def _push_scheduled(self, dt: datetime.datetime):
        while (next_scheduled_dt := self._scheduler_queue.peek_next_event_dt()) and next_scheduled_dt <= dt:
            next_scheduled_dt, job = self._scheduler_queue.pop()
            # Push scheduled job into the task pool for processing.
            await self._handlers_task_pool.push(self._execute_scheduled(next_scheduled_dt, job))

    async def _push_events(self, dt: datetime.datetime):
        # Pop events and feed the pool.
        for source, evnt in self._event_mux.pop_while(dt):
            # Check that events from the same source are returned in order.
            prev_event_dt = self._prev_event_dt.get(source)
            if prev_event_dt is not None and evnt.when < prev_event_dt:
                self.on_error(logs.StructuredMessage(
                    "Events returned out of order", source=type(source), previous=prev_event_dt, current=evnt.when
                ))
                # TODO: Not ignoring out-of-order events should be an option.
                continue
            self._prev_event_dt[source] = evnt.when

            # Push event into the task pool for processing.
            await self._handlers_task_pool.push(
                self._dispatch_event(EventDispatch(
                    event=evnt,
                    handlers=self._event_handlers.get(source, [])
                ))
            )


async def gather_no_raise(*awaitables):
    await asyncio.gather(*[await_no_raise(awaitable) for awaitable in awaitables])


async def await_no_raise(coro: Awaitable[Any], message: str = "Unhandled exception"):
    with helpers.no_raise(logger, message):
        await coro


def realtime_dispatcher(max_concurrent: int = 50) -> RealtimeDispatcher:
    """
    Creates an event dispatcher suitable for live trading.

    :param max_concurrent: The maximum number of events to process concurrently.
    """
    return RealtimeDispatcher(max_concurrent=max_concurrent)


def backtesting_dispatcher(max_concurrent: int = 50) -> BacktestingDispatcher:
    """
    Creates an event dispatcher suitable for backtesting.

    :param max_concurrent: The maximum number of events to process concurrently.
    """
    return BacktestingDispatcher(max_concurrent=max_concurrent)



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from typing import Any, Dict, List, Optional
import abc

from . import common, helpers, margin_requests
from .client import margin as margin_client
from basana.core.enums import OrderOperation
from basana.core.pair import Pair


class CanceledOCOOrder(common.CanceledOCOOrder):
    pass


class CanceledOrder(common.CanceledOrder):
    pass


class OpenOrder(common.OpenOrder):
    pass


class CreatedOCOOrder(common.CreatedOCOOrder):
    pass


class Fill(common.Fill):
    pass


class OCOOrderInfo(common.OCOOrderInfo):
    pass


class OrderInfo(common.OrderInfo):
    pass


class Balance(common.Balance):
    @property
    def borrowed(self) -> Decimal:
        """The borrowed balance."""
        return Decimal(self.json["borrowed"])


class Trade(common.Trade):
    @property
    def is_isolated(self) -> bool:
        return self.json["isIsolated"]


class CreatedOrder(common.CreatedOrder):
    @property
    def fills(self) -> List[Fill]:
        """The fills.

        Only available for FULL responses.
        """
        return [Fill(fill) for fill in self.json.get("fills", [])]


class Account(metaclass=abc.ABCMeta):
    @property
    @abc.abstractmethod
    def client(self) -> margin_client.MarginAccount:
        raise NotImplementedError()

    async def create_order(self, order_request: margin_requests.ExchangeOrder) -> CreatedOrder:
        created_order = await order_request.create_order(self.client)
        return CreatedOrder(created_order)

    async def create_market_order(
            self, operation: OrderOperation, pair: Pair, amount: Optional[Decimal] = None,
            quote_amount: Optional[Decimal] = None, client_order_id: Optional[str] = None,
            side_effect_type: str = "NO_SIDE_EFFECT", **kwargs: Dict[str, Any]
    ) -> CreatedOrder:
        """Creates a market order.

        Check https://binance-docs.github.io/apidocs/spot/en/#margin-account-new-order-trade for more information.
        If the order can't be created a :class:`basana.external.binance.exchange.Error` will be raised.

        :param operation: The order operation.
        :param pair: The pair to trade.
        :param amount: The amount to buy/sell in base units.
        :param quote_amount: The amount to buy/sell in quote units.
        :param client_order_id: A client order id.
        :param side_effect_type: One of NO_SIDE_EFFECT, MARGIN_BUY or AUTO_REPAY.
        :param kwargs: Additional keyword arguments that will be forwarded.

        .. note::

          * Either amount or quote_amount should be set, but not both.
        """

        return await self.create_order(margin_requests.MarketOrder(
            operation, pair, amount=amount, quote_amount=quote_amount, client_order_id=client_order_id,
            side_effect_type=side_effect_type, **kwargs
        ))

    async def create_limit_order(
            self, operation: OrderOperation, pair: Pair, amount: Decimal, limit_price: Decimal,
            side_effect_type: str = "NO_SIDE_EFFECT", time_in_force: str = "GTC", client_order_id: Optional[str] = None,
            **kwargs: Dict[str, Any]
    ) -> CreatedOrder:
        """Creates a limit order.

        Check https://binance-docs.github.io/apidocs/spot/en/#margin-account-new-order-trade for more information.
        If the order can't be created a :class:`basana.external.binance.exchange.Error` will be raised.

        :param operation: The order operation.
        :param pair: The pair to trade.
        :param amount: The amount to buy/sell in base units.
        :param limit_price: The limit price.
        :param side_effect_type: One of NO_SIDE_EFFECT, MARGIN_BUY or AUTO_REPAY.
        :param time_in_force: The time in force.
        :param client_order_id: A client order id.
        :param kwargs: Additional keyword arguments that will be forwarded.
        """

        return await self.create_order(margin_requests.LimitOrder(
            operation, pair, amount, limit_price, side_effect_type=side_effect_type, time_in_force=time_in_force,
            client_order_id=client_order_id, **kwargs
        ))

    async def create_stop_limit_order(
            self, operation: OrderOperation, pair: Pair, amount: Decimal, stop_price: Decimal, limit_price: Decimal,
            side_effect_type: str = "NO_SIDE_EFFECT", time_in_force: str = "GTC",
            client_order_id: Optional[str] = None, **kwargs: Dict[str, Any]
    ) -> CreatedOrder:
        """Creates a stop limit order.

        Check https://binance-docs.github.io/apidocs/spot/en/#margin-account-new-order-trade for more information.
        If the order can't be created a :class:`basana.external.binance.exchange.Error` will be raised.

        :param operation: The order operation.
        :param pair: The pair to trade.
        :param amount: The amount to buy/sell in base units.
        :param stop_price: The stop price.
        :param limit_price: The limit price.
        :param side_effect_type: One of NO_SIDE_EFFECT, MARGIN_BUY or AUTO_REPAY.
        :param time_in_force: The time in force.
        :param client_order_id: A client order id.
        :param kwargs: Additional keyword arguments that will be forwarded.
        """

        return await self.create_order(margin_requests.StopLimitOrder(
            operation, pair, amount, stop_price, limit_price, side_effect_type=side_effect_type,
            time_in_force=time_in_force, client_order_id=client_order_id, **kwargs
        ))

    async def get_order_info(
            self, pair: Pair, order_id: Optional[str] = None, client_order_id: Optional[str] = None,
            include_trades: bool = True
    ) -> OrderInfo:
        """Returns information about an order.

        :param pair: That trading pair.
        :param order_id: The order id.
        :param client_order_id: The client order id.

        .. note::

          * Either order_id or client_order_id should be set, but not both.
          * Including trades requires making an extra request to Binance.
        """
        order_book_symbol = helpers.pair_to_order_book_symbol(pair)
        order_info = await self.client.query_order(
            order_book_symbol, order_id=None if order_id is None else int(order_id),
            orig_client_order_id=client_order_id
        )
        trades = []
        if include_trades:
            trades = [
                Trade(trade) for trade in
                await self.client.get_trades(order_book_symbol, order_id=order_info["orderId"])
            ]
        return OrderInfo(order_info, trades)

    async def get_open_orders(self, pair: Optional[Pair] = None) -> List[OpenOrder]:
        """Returns open orders.

        :param pair: If set, only open orders matching this pair will be returned, otherwise all open orders will be
            returned.
        """
        order_book_symbol = None
        if pair:
            order_book_symbol = helpers.pair_to_order_book_symbol(pair)
        return [
            OpenOrder(open_order) for open_order in await self.client.get_open_orders(order_book_symbol)
        ]

    async def cancel_order(
            self, pair: Pair, order_id: Optional[str] = None, client_order_id: Optional[str] = None,
    ) -> CanceledOrder:
        """Cancels an order.

        If the order can't be canceled a :class:`basana.external.binance.exchange.Error` will be raised.

        :param pair: The trading pair.
        :param order_id: The order id.
        :param client_order_id: The client order id.

        .. note::

          * Either order_id or client_order_id should be set, but not both.
        """
        canceled_order = await self.client.cancel_order(
            helpers.pair_to_order_book_symbol(pair), order_id=None if order_id is None else int(order_id),
            orig_client_order_id=client_order_id
        )
        return CanceledOrder(canceled_order)

    async def create_oco_order(
            self, operation: OrderOperation, pair: Pair, amount: Decimal, limit_price: Decimal, stop_price: Decimal,
            stop_limit_price: Optional[Decimal] = None, side_effect_type: str = "NO_SIDE_EFFECT",
            stop_limit_time_in_force: str = "GTC", list_client_order_id: Optional[str] = None,
            limit_client_order_id: Optional[str] = None, stop_client_order_id: Optional[str] = None,
            **kwargs: Dict[str, Any]
    ) -> CreatedOCOOrder:
        """Creates an OCO order.

        Check https://binance-docs.github.io/apidocs/spot/en/#margin-account-new-oco-trade for more information.
        If the order can't be created a :class:`basana.external.binance.exchange.Error` will be raised.

        :param operation: The order operation.
        :param pair: The pair to trade.
        :param amount: The amount to buy/sell in base units.
        :param limit_price: The limit price.
        :param stop_price: The stop price.
        :param stop_limit_price: The stop limit price.
        :param side_effect_type: One of NO_SIDE_EFFECT, MARGIN_BUY or AUTO_REPAY.
        :param stop_limit_time_in_force: The time in force for the stop limit order.
        :param list_client_order_id: A client id for the order list.
        :param limit_client_order_id: A client id for the limit order.
        :param stop_client_order_id: A client id for the stop order.
        :param kwargs: Additional keyword arguments that will be forwarded.
        """
        order_req = margin_requests.OCOOrder(
            operation, pair, amount, limit_price, stop_price, stop_limit_price=stop_limit_price,
            side_effect_type=side_effect_type, stop_limit_time_in_force=stop_limit_time_in_force,
            list_client_order_id=list_client_order_id, limit_client_order_id=limit_client_order_id,
            stop_client_order_id=stop_client_order_id, **kwargs
        )
        created_order = await order_req.create_order(self.client)
        return CreatedOCOOrder(created_order)

    async def get_oco_order_info(
            self, order_list_id: Optional[str] = None, client_order_list_id: Optional[str] = None,
    ) -> OCOOrderInfo:
        """Returns information about an OCO order.

        :param order_list_id: The order list id.
        :param client_order_list_id: A client id for the order list.

        .. note::

          * Either order_list_id or client_order_list_id should be set, but not both.
        """
        order_info = await self.client.query_oco_order(
            order_list_id=None if order_list_id is None else int(order_list_id),
            client_order_list_id=client_order_list_id
        )
        return OCOOrderInfo(order_info)

    async def cancel_oco_order(
            self, pair: Pair, order_list_id: Optional[str] = None, client_order_list_id: Optional[str] = None,
    ) -> CanceledOCOOrder:
        """Cancels an OCO order.

        If the order can't be canceled a :class:`basana.external.binance.exchange.Error` will be raised.

        :param pair: The trading pair.
        :param order_list_id: The order list id.
        :param client_order_list_id: A client id for the order list.

        .. note::

          * Either order_list_id or client_order_list_id should be set, but not both.
        """
        canceled_order = await self.client.cancel_oco_order(
            helpers.pair_to_order_book_symbol(pair),
            order_list_id=None if order_list_id is None else int(order_list_id),
            client_order_list_id=client_order_list_id
        )
        return CanceledOCOOrder(canceled_order)



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from typing import Sequence, Tuple
import datetime

from dateutil import tz

from basana.core.event_sources import csv
from basana.core import pair, event, bar


######################################################################
# Yahoo Finance CSV parser
#
# File format:
# Date,Open,High,Low,Close,Volume,Adj Close
#
# The csv Date column must have the following format: YYYY-MM-DD


def adjust_ohlc(
        open: Decimal, high: Decimal, low: Decimal, close: Decimal, adj_close: Decimal
) -> Tuple[Decimal, Decimal, Decimal, Decimal]:
    adj_factor = adj_close / close
    open *= adj_factor
    high *= adj_factor
    low *= adj_factor
    close = adj_close
    return open, high, low, close


def sanitize_ohlc(open: Decimal, high: Decimal, low: Decimal, close: Decimal) -> Tuple[
    Decimal, Decimal, Decimal, Decimal
]:
    if low > open:
        low = open
    if low > close:
        low = close
    if high < open:
        high = open
    if high < close:
        high = close
    return open, high, low, close


class RowParser(csv.RowParser):
    def __init__(
            self, pair: pair.Pair, adjust_ohlc: bool = False, tzinfo: datetime.tzinfo = tz.tzlocal(),
            timedelta: datetime.timedelta = datetime.timedelta(hours=24)
    ):
        self.pair = pair
        self.tzinfo = tzinfo
        self.timedelta = timedelta
        self.sanitize = False
        self.adjust_ohlc = adjust_ohlc

    def parse_row(self, row_dict: dict) -> Sequence[event.Event]:
        dt = datetime.datetime.strptime(row_dict["Date"], "%Y-%m-%d").replace(tzinfo=self.tzinfo)
        open, high, low, close = (
            Decimal(row_dict["Open"]), Decimal(row_dict["High"]), Decimal(row_dict["Low"]), Decimal(row_dict["Close"])
        )
        if self.sanitize:
            open, high, low, close = sanitize_ohlc(open, high, low, close)
        if self.adjust_ohlc:
            open, high, low, close = adjust_ohlc(open, high, low, close, Decimal(row_dict["Adj Close"]))

        return [
            bar.BarEvent(
                dt + self.timedelta,
                bar.Bar(dt, self.pair, open, high, low, close, Decimal(row_dict["Volume"]))
            )
        ]


class CSVBarSource(csv.EventSource):
    def __init__(
            self, pair: pair.Pair, csv_path: str, adjust_ohlc: bool = False, sort: bool = True,
            tzinfo: datetime.tzinfo = tz.tzlocal(),
            timedelta: datetime.timedelta = datetime.timedelta(hours=24),
            dict_reader_kwargs: dict = {}
    ):
        self.row_parser = RowParser(pair, adjust_ohlc=adjust_ohlc, tzinfo=tzinfo, timedelta=timedelta)
        super().__init__(csv_path, self.row_parser, sort=sort, dict_reader_kwargs=dict_reader_kwargs)



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from typing import cast, Callable, Dict, List, Optional, Sequence, Tuple
import dataclasses
import logging
import uuid

from basana.backtesting import account_balances, config, errors, fees, lending, loan_mgr, liquidity, \
    orders, order_mgr, prices, requests
from basana.core import bar, dispatcher, enums, event, logs
from basana.core.pair import Pair, PairInfo
from basana.backtesting.lending import base as lending_base


logger = logging.getLogger(__name__)

BarEventHandler = bar.BarEventHandler
Error = errors.Error
LiquidityStrategyFactory = Callable[[], liquidity.LiquidityStrategy]
OrderEvent = order_mgr.OrderEvent
OrderEventHandler = order_mgr.OrderEventHandler
OrderInfo = orders.OrderInfo
OrderOperation = enums.OrderOperation


@dataclasses.dataclass
class Balance:
    #: The available balance.
    available: Decimal
    #: The total balance (available + hold - borrowed).
    total: Decimal = dataclasses.field(init=False)
    #: The balance on hold (reserved for open sell orders).
    hold: Decimal
    #: The balance borrowed.
    borrowed: Decimal

    def __post_init__(self):
        self.total = self.available + self.hold - self.borrowed


@dataclasses.dataclass
class CreatedOrder:
    #: The order id.
    id: str


@dataclasses.dataclass
class CanceledOrder:
    #: The order id.
    id: str


@dataclasses.dataclass
class OpenOrder:
    #: The order id.
    id: str
    #: The operation.
    operation: OrderOperation
    #: The original amount.
    amount: Decimal
    #: The amount filled.
    amount_filled: Decimal


class Exchange:
    """
    This class implements a backtesting exchange.

    This backtesting exchange has support for Market, Limit, Stop and Stop Limit orders and it will simulate order
    execution based on summarized trading activity (:class:`basana.BarEvent`).

    :param dispatcher: The event dispatcher.
    :param initial_balances: The initial balance for each currency/symbol/etc.
    :param liquidity_strategy_factory: A callable that returns a new liquidity strategy.
    :param fee_strategy: The stragegy to use to calculate fees.
    :param default_pair_info: The default pair information if a specific one was not set using
        :meth:`Exchange.set_pair_info`.
    :param bid_ask_spread: The spread to use for :meth:`Exchange.get_bid_ask`.
    :param lending_strategy: The strategy to use for managing loans.
    """
    def __init__(
            self,
            dispatcher: dispatcher.BacktestingDispatcher,
            initial_balances: Dict[str, Decimal],
            liquidity_strategy_factory: LiquidityStrategyFactory = liquidity.VolumeShareImpact,
            fee_strategy: fees.FeeStrategy = fees.NoFee(),
            default_pair_info: Optional[PairInfo] = PairInfo(base_precision=0, quote_precision=2),
            bid_ask_spread: Decimal = Decimal("0.5"),
            lending_strategy: lending.LendingStrategy = lending.NoLoans()
    ):
        self._dispatcher = dispatcher
        self._balances = account_balances.AccountBalances(initial_balances)
        self._bar_event_source: Dict[Pair, event.FifoQueueEventSource] = {}
        self._config = config.Config(None, default_pair_info)
        self._prices = prices.Prices(bid_ask_spread, self._config)
        self._loan_mgr = loan_mgr.LoanManager(
            lending_strategy,
            lending_base.ExchangeContext(
                dispatcher=dispatcher,
                account_balances=self._balances,
                prices=self._prices,
                config=self._config
            )
        )
        self._order_mgr = order_mgr.OrderManager(
            order_mgr.ExchangeContext(
                dispatcher=dispatcher, account_balances=self._balances, prices=self._prices,
                fee_strategy=fee_strategy, liquidity_strategy_factory=liquidity_strategy_factory,
                loan_mgr=self._loan_mgr, config=self._config
            )
        )

    async def get_balance(self, symbol: str) -> Balance:
        """
        Returns the balance for a specific currency/symbol/etc..

        :param symbol: The currency/symbol/etc..
        """
        return self._get_balance(symbol)

    async def get_balances(self) -> Dict[str, Balance]:
        """
        Returns all balances.
        """
        ret = {}
        for symbol in self._balances.get_symbols():
            ret[symbol] = self._get_balance(symbol)
        return ret

    async def get_bid_ask(self, pair: Pair) -> Tuple[Decimal, Decimal]:
        """
        Returns the last bid and ask price.

        This is calculated using the closing price of the last bar, and the bid/ask spread specified during
        initialization.

        :param pair: The trading pair.
        """
        return self._prices.get_bid_ask(pair)

    async def create_order(self, order_request: requests.ExchangeOrder) -> CreatedOrder:
        # Validate request parameters.
        pair_info = await self.get_pair_info(order_request.pair)
        order_request.validate(pair_info)

        order = order_request.create_order(uuid.uuid4().hex)
        self._order_mgr.add_order(order)
        logger.debug(logs.StructuredMessage("Request accepted", order_id=order.id))
        return CreatedOrder(id=order.id)

    async def create_market_order(
            self, operation: OrderOperation, pair: Pair, amount: Decimal, auto_borrow: bool = False,
            auto_repay: bool = False
    ) -> CreatedOrder:
        """
        Creates a market order.

        A market order is an order to immediately buy or sell at the best available price.
        Generally, this type of order will be executed on the next bar using the open price as a reference, and
        according to the rules defined by the liquidity strategy.
        If the order is not filled on the next bar, due to lack of liquidity or funds, the order will be canceled.

        If the order can't be created an :class:`Error` will be raised.

        :param operation: The order operation.
        :param pair: The pair to trade.
        :param amount: The base amount to buy/sell.
        :param auto_borrow: Automatically borrow missing funds.
        :param auto_repay: Automatically repay open loans once the order gets filled.
        """
        return await self.create_order(requests.MarketOrder(
            operation, pair, amount, auto_borrow=auto_borrow, auto_repay=auto_repay
        ))

    async def create_limit_order(
            self, operation: OrderOperation, pair: Pair, amount: Decimal, limit_price: Decimal,
            auto_borrow: bool = False, auto_repay: bool = False
    ) -> CreatedOrder:
        """
        Creates a limit order.

        A limit order is an order to buy or sell at a specific price or better.
        A buy limit order can only be executed at the limit price or lower, and a sell limit order can only be executed
        at the limit price or higher.

        If the order can't be created an :class:`Error` will be raised.

        :param operation: The order operation.
        :param pair: The pair to trade.
        :param amount: The base amount to buy/sell.
        :param limit_price: The limit price.
        :param auto_borrow: Automatically borrow missing funds.
        :param auto_repay: Automatically repay open loans once the order gets filled.
        """
        return await self.create_order(requests.LimitOrder(
            operation, pair, amount, limit_price, auto_borrow=auto_borrow, auto_repay=auto_repay
        ))

    async def create_stop_order(
            self, operation: OrderOperation, pair: Pair, amount: Decimal, stop_price: Decimal,
            auto_borrow: bool = False, auto_repay: bool = False
    ) -> CreatedOrder:
        """
        Creates a stop order.

        A stop order, also referred to as a stop-loss order, is an order to buy or sell once the price reaches a
        specified price, known as the stop price.
        When the stop price is reached, a stop order becomes a market order.

        * A buy stop order is entered at a stop price above the current market price. Investors generally use a buy
          stop order to limit a loss or to protect a profit on an instrument that they have sold short.
        * A sell stop order is entered at a stop price below the current market price. Investors generally use a sell
          stop order to limit a loss or to protect a profit on an instrument that they own.

        If the order can't be created an :class:`Error` will be raised.

        :param operation: The order operation.
        :param pair: The pair to trade.
        :param amount: The base amount to buy/sell.
        :param stop_price: The stop price.
        :param auto_borrow: Automatically borrow missing funds.
        :param auto_repay: Automatically repay open loans once the order gets filled.
        """
        return await self.create_order(requests.StopOrder(
            operation, pair, amount, stop_price, auto_borrow=auto_borrow, auto_repay=auto_repay
        ))

    async def create_stop_limit_order(
            self, operation: OrderOperation, pair: Pair, amount: Decimal, stop_price: Decimal, limit_price: Decimal,
            auto_borrow: bool = False, auto_repay: bool = False
    ) -> CreatedOrder:
        """
        Creates a stop limit order.

        A stop-limit order is an order to buy or sell that combines the features of a stop order and a limit order.
        Once the stop price is reached, a stop-limit order becomes a limit order that will be executed at a specified
        price (or better).

        If the order can't be created an :class:`Error` will be raised.

        :param operation: The order operation.
        :param pair: The pair to trade.
        :param amount: The base amount to buy/sell.
        :param stop_price: The stop price.
        :param limit_price: The limit price.
        :param auto_borrow: Automatically borrow missing funds.
        :param auto_repay: Automatically repay open loans once the order gets filled.
        """
        return await self.create_order(requests.StopLimitOrder(
            operation, pair, amount, stop_price, limit_price, auto_borrow=auto_borrow, auto_repay=auto_repay
        ))

    async def cancel_order(self, order_id: str) -> CanceledOrder:
        """
        Cancels an order.

        If the order doesn't exist, or its not open, an :class:`Error` will be raised.

        :param order_id: The order id.
        """
        self._order_mgr.cancel_order(order_id)
        return CanceledOrder(id=order_id)

    async def get_order_info(self, order_id: str) -> OrderInfo:
        """
        Returns information about an order.

        If the order doesn't exist, or its not open, an :class:`Error` will be raised.

        :param order_id: The order id.
        """
        order = self._order_mgr.get_order(order_id)
        if not order:
            raise errors.NotFound("Order not found")
        return order.get_order_info()

    async def get_open_orders(self, pair: Optional[Pair] = None) -> List[OpenOrder]:
        """
        Returns open orders.

        :param pair: If set, only open orders matching this pair will be returned, otherwise all open orders will be
            returned.
        """
        return [
            OpenOrder(
                id=order.id,
                operation=order.operation,
                amount=order.amount,
                amount_filled=order.amount_filled
            )
            for order in self._order_mgr.get_open_orders()
            if pair is None or order.pair == pair
        ]

    async def get_orders(self, pair: Optional[Pair] = None, is_open: Optional[bool] = None) -> List[OrderInfo]:
        """
        Returns orders filtered by various conditions.

        :param pair: If set, only orders matching this pair will be returned.
        :param is_open: If set, only open or closed orders will be returned.
        """

        orders = self._order_mgr.get_all_orders()
        if pair:
            orders = filter(lambda order: order.pair == pair, orders)
        if is_open is not None:
            orders = filter(lambda order: order.is_open == is_open, orders)
        return [order.get_order_info() for order in orders]

    def add_bar_source(self, bar_source: event.EventSource):
        """
        Adds an event source that produces :class:`basana.BarEvent` instances.

        These will be used to drive the backtest.

        :param bar_source: An event source that produces :class:`basana.BarEvent` instances.
        """
        self._dispatcher.subscribe(bar_source, self._on_bar_event)

    def subscribe_to_bar_events(self, pair: Pair, event_handler: BarEventHandler):
        """
        Registers an async callable that will be called when a new bar is available.

        :param pair: The trading pair.
        :param event_handler: An async callable that receives a basana.BarEvent.
        """
        # Get/create the event source for the given pair.
        event_source = self._bar_event_source.get(pair)
        if event_source is None:
            event_source = event.FifoQueueEventSource()
            self._bar_event_source[pair] = event_source
        self._dispatcher.subscribe(event_source, cast(dispatcher.EventHandler, event_handler))

    def subscribe_to_order_events(self, event_handler: OrderEventHandler):
        """
        Registers an async callable that will be called when an order is accepted or updated.

        :param event_handler: The event handler.
        """
        self._order_mgr.subscribe_to_order_events(event_handler)

    async def get_pair_info(self, pair: Pair) -> PairInfo:
        """
        Returns information about a trading pair.

        :param pair: The trading pair.
        """
        return self._get_pair_info(pair)

    def set_pair_info(self, pair: Pair, pair_info: PairInfo):
        """
        Set information about a trading pair.

        :param pair: The trading pair.
        :param pair_info: The pair information.
        """
        self._config.set_pair_info(pair, pair_info)

    def set_symbol_precision(self, symbol: str, precision: int):
        """
        Set precision for a symbol.

        :param symbol: The symbol.
        :param precision: The precision.
        """
        self._config.set_symbol_info(symbol, config.SymbolInfo(precision=precision))

    async def create_loan(self, symbol: str, amount: Decimal) -> lending.LoanInfo:
        """
        Creates a loan.

        :param symbol: The symbol to borrow.
        :param amount: The amount to borrow.
        """

        return self._loan_mgr.create_loan(symbol, amount)

    async def get_loans(
            self, borrowed_symbol: Optional[str] = None, is_open: Optional[bool] = None
    ) -> List[lending.LoanInfo]:
        """
        Returns loans filtered by various conditions.

        :param borrowed_symbol: If set, only loans matching this borrowed symbol will be returned.
        :param is_open: If set, only open or closed loans will be returned.
        """
        return self._loan_mgr.get_loans(borrowed_symbol=borrowed_symbol, is_open=is_open)

    async def get_loan(self, loan_id: str) -> lending.LoanInfo:
        """
        Returns information about a loan.

        :param loan_id: The loan id.
        """
        loan_info = self._loan_mgr.get_loan(loan_id)
        if not loan_info:
            raise errors.NotFound("Loan not found")
        return loan_info

    async def repay_loan(self, loan_id: str):
        """
        Repays a loan fully.

        :param loan_id: The loan id.
        """
        return self._loan_mgr.repay_loan(loan_id)

    def _get_pair_info(self, pair: Pair) -> PairInfo:
        return self._config.get_pair_info(pair)

    async def _on_bar_event(self, event: event.Event):
        assert isinstance(event, bar.BarEvent), f"{event} is not an instance of bar.BarEvent"

        self._prices.on_bar_event(event)
        self._order_mgr.on_bar_event(event)

        # Forward the event if necessary.
        event_source = self._bar_event_source.get(event.bar.pair)
        if event_source:
            event_source.push(event)

    def _get_all_orders(self) -> Sequence[orders.Order]:
        return list(self._order_mgr.get_all_orders())

    def _get_dispatcher(self) -> dispatcher.BacktestingDispatcher:
        return self._dispatcher

    def _get_balance(self, symbol: str) -> Balance:
        available = self._balances.get_available_balance(symbol)
        hold = self._balances.get_balance_on_hold(symbol)
        borrowed = self._balances.get_borrowed_balance(symbol)
        return Balance(
            available=available, hold=hold, borrowed=borrowed
        )



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from typing import Optional, Dict
import datetime

from . import client, config, helpers, margin, user_data, websockets, websocket_mgr
from .client import margin as margin_client
from basana.core.config import get_config_value
from basana.core.pair import Pair
import basana as bs


# Forward declarations
OrderEvent = user_data.OrderEvent
OrderEventHandler = user_data.OrderEventHandler
OrderUpdate = user_data.OrderUpdate
UserDataEvent = user_data.Event
UserDataEventHandler = user_data.UserDataEventHandler


class IsolatedBalance:
    def __init__(self, json: dict):
        self.json = json

    @property
    def base_asset(self) -> str:
        """The base asset."""
        return self.json["baseAsset"]["asset"]

    @property
    def base_asset_balance(self) -> margin.Balance:
        """The base asset balance."""
        return margin.Balance(self.json["baseAsset"])

    @property
    def quote_asset(self) -> str:
        """The quote asset."""
        return self.json["quoteAsset"]["asset"]

    @property
    def quote_asset_balance(self) -> margin.Balance:
        """The quote asset balance."""
        return margin.Balance(self.json["quoteAsset"])


class IsolatedMarginUserDataChannel(websockets.Channel):
    def __init__(self, pair: bs.Pair):
        self._pair = pair
        self._listen_key = None

    @property
    def alias(self) -> str:
        symbol = helpers.pair_to_order_book_symbol(self._pair)
        return f"isolated_margin_user_data_{symbol.lower()}"

    @property
    def stream(self) -> str:
        assert self._listen_key, "resolve_stream_name not called"
        return self._listen_key

    async def resolve_stream_name(self, api_client: client.APIClient):
        symbol = helpers.pair_to_order_book_symbol(self._pair)
        self._listen_key = (await api_client.isolated_margin_account.create_listen_key(symbol))["listenKey"]

    def keep_alive_period(self, config_overrides: dict = {}) -> Optional[datetime.timedelta]:
        return datetime.timedelta(
            seconds=get_config_value(
                config.DEFAULTS, "api.websockets.isolated_margin.user_data_stream.heartbeat", overrides=config_overrides
            )
        )

    async def keep_alive(self, api_client: client.APIClient):
        assert self._listen_key, "resolve_stream_name not called"
        symbol = helpers.pair_to_order_book_symbol(self._pair)
        await api_client.isolated_margin_account.keep_alive_listen_key(symbol, self._listen_key)


class Account(margin.Account):
    """Isolated margin account."""
    def __init__(self, cli: margin_client.IsolatedMarginAccount, ws_mgr: websocket_mgr.WebsocketManager):
        self._cli = cli
        self._ws_mgr = ws_mgr

    @property
    def client(self) -> margin_client.IsolatedMarginAccount:
        return self._cli

    async def get_balances(self) -> Dict[Pair, IsolatedBalance]:
        """Returns all balances."""
        account_info = await self.client.get_account_information()
        ret = {}
        for isolated_balance in account_info["assets"]:
            isolated_balance = IsolatedBalance(isolated_balance)
            pair = Pair(isolated_balance.base_asset, isolated_balance.quote_asset)
            ret[pair] = isolated_balance
        return ret

    async def transfer_from_spot_account(self, asset: str, pair: Pair, amount: Decimal) -> dict:
        """Transfer balances from the spot account to the isolated margin account.

        If the transfer can't be completed a :class:`basana.external.binance.exchange.Error` will be raised.

        :param asset: The asset to transfer.
        :param pair: The trading pair.
        :param amount: The amount to transfer.
        """
        return await self.client.transfer_from_spot_account(asset, helpers.pair_to_order_book_symbol(pair), amount)

    async def transfer_to_spot_account(self, asset: str, pair: Pair, amount: Decimal) -> dict:
        """Transfer balances from the isolated margin account to the spot account.

        If the transfer can't be completed a :class:`basana.external.binance.exchange.Error` will be raised.

        :param asset: The asset to transfer.
        :param pair: The trading pair.
        :param amount: The amount to transfer.
        """
        return await self.client.transfer_to_spot_account(asset, helpers.pair_to_order_book_symbol(pair), amount)

    def subscribe_to_user_data_events(self, pair: Pair, event_handler: UserDataEventHandler):
        """
        Registers an async callable that will be called for every new user data event.

        Works as defined in https://developers.binance.com/docs/margin_trading/trade-data-stream.

        :param pair: The trading pair.
        :param event_handler: The event handler.
        """

        self._ws_mgr.subscribe_to_user_data_events(
            IsolatedMarginUserDataChannel(pair),
            lambda ws_cli: user_data.WebSocketEventSource(ws_cli),
            event_handler
        )

    def subscribe_to_order_events(self, pair: Pair, event_handler: OrderEventHandler):
        """
        Registers an async callable that will be called for every new order update.

        Works as defined in https://developers.binance.com/docs/margin_trading/trade-data-stream/Event-Order-Update.

        :param pair: The trading pair.
        :param event_handler: The event handler.
        """

        self._ws_mgr.subscribe_to_order_events(
            IsolatedMarginUserDataChannel(pair),
            lambda ws_cli: user_data.WebSocketEventSource(ws_cli),
            event_handler
        )



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from typing import Any, Awaitable, Callable, List, Optional, Tuple
import asyncio
import datetime
import logging

from basana.core import dt, event, logs, pair


logger = logging.getLogger(__name__)


class InvalidBar(Exception):
    pass


class Bar:
    """A Bar, also known as candlestick or OHLC, is the summary of the trading activity in a given period.

    :param datetime: The beginning of the period. It must have timezone information set.
    :param pair: The trading pair.
    :param open: The opening price.
    :param high: The highest traded price.
    :param low: The lowest traded price.
    :param close: The closing price.
    :param volume: The volume traded.
    """

    def __init__(
            self, datetime: datetime.datetime, pair: pair.Pair,
            open: Decimal, high: Decimal, low: Decimal, close: Decimal, volume: Decimal
    ):
        if high < low:
            raise InvalidBar(f"high < low on {datetime}")
        elif high < open:
            raise InvalidBar(f"high < open on {datetime}")
        elif high < close:
            raise InvalidBar(f"high < close on {datetime}")
        elif low > open:
            raise InvalidBar(f"low > open on {datetime}")
        elif low > close:
            raise InvalidBar(f"low > close on {datetime}")

        #: The beginning of the period.
        self.datetime = datetime
        #: The trading pair.
        self.pair = pair
        #: The opening price.
        self.open = open
        #: The highest traded price.
        self.high = high
        #: The lowest traded price.
        self.low = low
        #: The closing price.
        self.close = close
        #: The volume traded.
        self.volume = volume


class BarEvent(event.Event):
    """An event for :class:`Bar` instances.

    :param when: The datetime when the event occurred. It must have timezone information set.
    :param bar: The bar.
    """

    def __init__(self, when: datetime.datetime, bar: Bar):
        super().__init__(when)

        #: The bar.
        self.bar = bar


class RealTimeTradesToBar(event.FifoQueueEventSource, event.Producer):
    def __init__(self, pair: pair.Pair, bar_duration: int, skip_first_bar: bool = True, flush_delay: float = 0.5):
        assert bar_duration > 0
        assert flush_delay >= 0
        super().__init__(producer=self)
        self._pair = pair
        self._bar_duration = bar_duration
        self._trades: List[Tuple[datetime.datetime, Decimal, Decimal]] = []
        self._skip_first_bar = skip_first_bar
        self._next_trade_ge: Optional[datetime.datetime] = None
        self._flush_delay = flush_delay

    def on_error(self, error: Any):
        logger.error(error)

    def push_trade(self, when: datetime.datetime, price: Decimal, amount: Decimal):
        # Trades must arrive in order.
        if self._next_trade_ge and when < self._next_trade_ge:
            self.on_error(logs.StructuredMessage(
                "Trade pushed out of order", last=self._next_trade_ge, current=when, pair=self._pair
            ))
            return

        self._trades.append((when, price, amount))
        self._next_trade_ge = when

    def _flush(self, begin: datetime.datetime, end: datetime.datetime):
        logger.debug(logs.StructuredMessage("Flushing", begin=begin, end=end, pair=self._pair))
        assert end > begin

        self._next_trade_ge = end if self._next_trade_ge is None else max(self._next_trade_ge, end)
        open = Decimal(0)
        high = Decimal(0)
        low = Decimal(0)
        close = Decimal(0)
        volume = Decimal(0)
        future_trades_begin = None

        # Calculate open, high, low, close and volume in the given window.
        for i, (when, price, amount) in enumerate(self._trades):
            if when < begin:
                self.on_error(logs.StructuredMessage(
                    "Trade is out of order", datetime=when, begin=begin, end=end, pair=self._pair
                ))
                continue
            # If the trade belongs to a future window, then we're done processing the current window.
            if when > end:
                future_trades_begin = i
                break

            open = price if not open else open
            high = price if not high else max(high, price)
            low = price if not low else min(low, price)
            close = price
            volume += amount

        # If there were trades in the current window then build the bar and publish the event.
        if volume and not self._skip_first_bar:
            bar = Bar(begin, self._pair, open, high, low, close, volume)
            self.push(BarEvent(end, bar))
        self._skip_first_bar = False

        # Dump the trades that were already processed
        self._trades = [] if future_trades_begin is None else self._trades[future_trades_begin:]

    async def main(self):
        now = dt.utc_now()
        begin = now - datetime.timedelta(seconds=now.timestamp() % self._bar_duration)
        end = begin + datetime.timedelta(seconds=self._bar_duration, milliseconds=-1)
        while True:
            sleep_time = (end - dt.utc_now()).total_seconds() + self._flush_delay
            if sleep_time > 0:
                await asyncio.sleep(sleep_time)
            self._flush(begin, end)
            begin += datetime.timedelta(seconds=self._bar_duration)
            end += datetime.timedelta(seconds=self._bar_duration)


BarEventHandler = Callable[[BarEvent], Awaitable[Any]]



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
import logging

from . import helpers
from basana.core import bar, event, websockets as core_ws
from basana.core.pair import Pair


logger = logging.getLogger(__name__)


class Bar(bar.Bar):
    def __init__(self, pair: Pair, json: dict):
        super().__init__(
            helpers.timestamp_to_datetime(int(json["t"])), pair, Decimal(json["o"]), Decimal(json["h"]),
            Decimal(json["l"]), Decimal(json["c"]), Decimal(json["v"])
        )
        self.pair: Pair = pair
        self.json: dict = json


# Generate BarEvents events from websocket messages.
class WebSocketEventSource(core_ws.ChannelEventSource):
    def __init__(self, pair: Pair, producer: event.Producer):
        super().__init__(producer=producer)
        self._pair: Pair = pair

    async def push_from_message(self, message: dict):
        kline_event = message["data"]
        kline = kline_event["k"]
        # Wait for the last update to the kline.
        if kline["x"] is False:
            return
        self.push(bar.BarEvent(
            helpers.timestamp_to_datetime(int(kline_event["E"])),
            Bar(self._pair, kline)
        ))


def get_channel(pair: Pair, interval: str) -> str:
    return "{}@kline_{}".format(helpers.pair_to_order_book_symbol(pair).lower(), interval)



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from typing import Any, Dict, Optional
import abc

from . import client, helpers
from basana.core.enums import OrderOperation
from basana.core.pair import Pair


class ExchangeOrder(metaclass=abc.ABCMeta):
    def __init__(
            self, operation: OrderOperation, pair: Pair, amount: Decimal, client_order_id: Optional[str] = None,
            **kwargs: Dict[str, Any]
    ):
        self._operation = operation
        self._pair = pair
        self._amount = amount
        self._client_order_id = client_order_id
        self._kwargs = kwargs

    def _get_action(self) -> str:
        return {
            OrderOperation.BUY: "buy",
            OrderOperation.SELL: "sell",
        }[self.operation]

    @property
    def pair(self) -> Pair:
        return self._pair

    @property
    def amount(self) -> Decimal:
        return self._amount

    @property
    def operation(self) -> OrderOperation:
        return self._operation

    @abc.abstractmethod
    async def create_order(self, cli: client.APIClient) -> dict:
        raise NotImplementedError()


class MarketOrder(ExchangeOrder):
    """
    Market order request.

    A market order is an order to buy or sell a stock at the best available price.
    Generally, this type of order will be executed immediately. However, the price at which a market order will be
    executed is not guaranteed.
    """

    def __init__(
            self, operation: OrderOperation, pair: Pair, amount: Decimal, client_order_id: Optional[str] = None,
            **kwargs: Dict[str, Any]
    ):
        super().__init__(operation, pair, amount, client_order_id=client_order_id, **kwargs)

    async def create_order(self, cli: client.APIClient) -> dict:
        return await cli.create_market_order(
            self._get_action(), helpers.pair_to_currency_pair(self.pair), self.amount,
            client_order_id=self._client_order_id, **self._kwargs
        )


class LimitOrder(ExchangeOrder):
    """Limit order request.

    A limit order is an order to buy or sell a stock at a specific price or better.
    A buy limit order can only be executed at the limit price or lower, and a sell limit order can only be executed
    at the limit price or higher.
    """

    def __init__(
            self, operation: OrderOperation, pair: Pair, amount: Decimal, limit_price: Decimal,
            client_order_id: Optional[str] = None, **kwargs: Dict[str, Any]
    ):
        super().__init__(operation, pair, amount, client_order_id=client_order_id, **kwargs)
        self._limit_price = limit_price

    @property
    def limit_price(self) -> Decimal:
        return self._limit_price

    async def create_order(self, cli: client.APIClient) -> dict:
        return await cli.create_limit_order(
            self._get_action(), helpers.pair_to_currency_pair(self.pair), self.amount, self.limit_price,
            client_order_id=self._client_order_id, **self._kwargs
        )


class InstantOrder(ExchangeOrder):
    def __init__(
            self, operation: OrderOperation, pair: Pair, amount: Decimal, amount_in_counter: bool = False,
            client_order_id: Optional[str] = None, **kwargs: Dict[str, Any]
    ):
        super().__init__(operation, pair, amount, client_order_id=client_order_id, **kwargs)
        self._amount_in_counter = amount_in_counter

    async def create_order(self, cli: client.APIClient) -> dict:
        return await cli.create_instant_order(
            self._get_action(), helpers.pair_to_currency_pair(self.pair), self.amount,
            amount_in_counter=self._amount_in_counter, client_order_id=self._client_order_id, **self._kwargs
        )



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum


# Using these values to avoid misunderstandings with external API values.
@enum.unique
class OrderOperation(enum.Enum):
    """Enumeration for order operations."""

    #:
    BUY = 100
    #:
    SELL = 101

    def __str__(self):
        return {
            OrderOperation.BUY: "buy",
            OrderOperation.SELL: "sell",
        }[self]


@enum.unique
class Position(enum.Enum):
    """Enumeration for positions."""

    #:
    LONG = 200
    #:
    SHORT = 201
    #:
    NEUTRAL = 202

    def __str__(self):
        return {
            Position.LONG: "long",
            Position.SHORT: "short",
            Position.NEUTRAL: "neutral",
        }[self]



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Optional

import aiohttp

from . import base, margin, spot
from basana.core import token_bucket


Error = base.Error


class APIClient:
    def __init__(
            self, api_key: Optional[str] = None, api_secret: Optional[str] = None,
            session: Optional[aiohttp.ClientSession] = None, tb: Optional[token_bucket.TokenBucketLimiter] = None,
            config_overrides: dict = {}
    ):
        self._client = base.BaseClient(
            api_key=api_key, api_secret=api_secret, session=session, tb=tb, config_overrides=config_overrides
        )

    async def get_exchange_info(self, symbol: Optional[str] = None) -> dict:
        params = {}
        if symbol:
            params["symbol"] = symbol
        return await self._client.make_request("GET", "/api/v3/exchangeInfo", qs_params=params)

    @property
    def spot_account(self) -> spot.SpotAccount:
        return spot.SpotAccount(self._client)

    @property
    def cross_margin_account(self) -> margin.CrossMarginAccount:
        return margin.CrossMarginAccount(self._client)

    @property
    def isolated_margin_account(self) -> margin.IsolatedMarginAccount:
        return margin.IsolatedMarginAccount(self._client)

    async def get_order_book(self, symbol: str, limit: Optional[int] = None) -> dict:
        params: Dict[str, Any] = {"symbol": symbol}
        if limit is not None:
            params["limit"] = limit
        return await self._client.make_request("GET", "/api/v3/depth", qs_params=params)

    async def get_candlestick_data(
            self, symbol: str, interval: str, start_time: Optional[int] = None, end_time: Optional[int] = None,
            limit: Optional[int] = None
    ) -> list:
        params: Dict[str, Any] = {
            "symbol": symbol,
            "interval": interval,
        }
        base.set_optional_params(params, (
            ("startTime", start_time),
            ("endTime", end_time),
            ("limit", limit),
        ))
        return await self._client.make_request("GET", "/api/v3/klines", qs_params=params)



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Optional
import dataclasses

from basana.backtesting import errors
from basana.core.pair import Pair, PairInfo


@dataclasses.dataclass(frozen=True)
class SymbolInfo:
    precision: int


class Config:
    def __init__(self, default_symbol_info: Optional[SymbolInfo] = None, default_pair_info: Optional[PairInfo] = None):
        self._symbol_info: Dict[str, SymbolInfo] = {}
        self._default_symbol_info = default_symbol_info
        self._pair_info: Dict[Pair, PairInfo] = {}
        self._default_pair_info = default_pair_info

    def set_pair_info(self, pair: Pair, pair_info: PairInfo):
        self._pair_info[pair] = pair_info

    def get_pair_info(self, pair: Pair) -> PairInfo:
        ret = self._pair_info.get(pair)

        # If we don't have config for this specific pair we'll try to build it using the config for the individual
        # symbols.
        if ret is None:
            base_symbol_config = self._symbol_info.get(pair.base_symbol)
            quote_symbol_config = self._symbol_info.get(pair.quote_symbol)
            if base_symbol_config and quote_symbol_config:
                ret = PairInfo(
                    base_precision=base_symbol_config.precision, quote_precision=quote_symbol_config.precision
                )
        # Default pair info, if set, is the last option.
        if ret is None:
            ret = self._default_pair_info

        if ret is None:
            raise errors.Error(f"No config for {pair}")
        return ret

    def set_symbol_info(self, symbol: str, symbol_info: SymbolInfo):
        self._symbol_info[symbol] = symbol_info

    def get_symbol_info(self, symbol: str) -> SymbolInfo:
        ret = self._symbol_info.get(symbol, self._default_symbol_info)
        if ret is None:
            raise errors.Error(f"No config for {symbol}")
        return ret



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import calendar
import datetime

from dateutil import tz


def is_naive(dt: datetime.datetime) -> bool:
    """Returns True if datetime is naive."""
    return dt.tzinfo is None or dt.tzinfo.utcoffset(dt) is None


def utc_now() -> datetime.datetime:
    """Returns the current datetime in UTC timezone."""
    return datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc)


def local_datetime(*args, **kwargs) -> datetime.datetime:
    return datetime.datetime(*args, **kwargs).replace(tzinfo=tz.tzlocal())


def local_now() -> datetime.datetime:
    """Returns the current datetime in local timezone."""
    return datetime.datetime.now().replace(tzinfo=tz.tzlocal())


def to_utc_timestamp(dt: datetime.datetime) -> int:
    # return (dt - datetime.datetime(1970, 1, 1).replace(tzinfo=datetime.timezone.utc)).total_seconds()
    return calendar.timegm(dt.utctimetuple())



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict
from urllib.parse import urlencode
import hashlib
import hmac
import time
import uuid

from basana.core import pair
from basana.core.enums import OrderOperation


def generate_nonce() -> str:
    return str(uuid.uuid4())


def get_auth_headers(
        hostname: str, api_key: str, api_secret: str, nonce: str, method: str, path: str,
        qs_params: Dict[str, str] = {}, data: Dict[str, str] = {}
) -> Dict[str, str]:
    assert path[0] == "/", "Leading slash is missing from path"

    timestamp = str(int(round(time.time() * 1000)))
    headers = {
        "X-Auth": "BITSTAMP {}".format(api_key),
        "X-Auth-Nonce": nonce,
        "X-Auth-Timestamp": timestamp,
        "X-Auth-Version": "v2",
    }
    if data:
        headers["Content-Type"] = "application/x-www-form-urlencoded"

    message = headers["X-Auth"]
    message += method.upper()
    message += hostname
    message += path
    if qs_params:  # pragma: no cover
        message += urlencode(qs_params)
    message += headers.get("Content-Type", "")
    message += headers["X-Auth-Nonce"]
    message += headers["X-Auth-Timestamp"]
    message += headers["X-Auth-Version"]
    if data:
        message += urlencode(data)

    headers["X-Auth-Signature"] = hmac.new(
        api_secret.encode(), msg=message.encode("utf-8"), digestmod=hashlib.sha256
    ).hexdigest()
    return headers


def pair_to_currency_pair(pair: pair.Pair) -> str:
    return "{}{}".format(pair.base_symbol.lower(), pair.quote_symbol.lower())


def order_type_to_order_operation(order_type: int) -> OrderOperation:
    ret = {
        0: OrderOperation.BUY,
        1: OrderOperation.SELL,
    }.get(order_type)
    assert ret is not None, f"Invalid order type {order_type}"
    return ret



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from typing import Dict, Generator, Generic, Iterable, List, Optional, Protocol, TypeVar

from basana.core.enums import OrderOperation


def get_base_sign_for_operation(operation: OrderOperation) -> Decimal:
    if operation == OrderOperation.BUY:
        base_sign = Decimal(1)
    else:
        assert operation == OrderOperation.SELL
        base_sign = Decimal(-1)
    return base_sign


class ExchangeObjectProto(Protocol):
    @property
    def id(self) -> str:  # pragma: no cover
        ...

    @property
    def is_open(self) -> bool:  # pragma: no cover
        ...


TExchangeObject = TypeVar('TExchangeObject', bound=ExchangeObjectProto)


class ExchangeObjectContainer(Generic[TExchangeObject]):
    def __init__(self):
        self._items: Dict[str, TExchangeObject] = {}  # Items by id.
        self._open_items: List[TExchangeObject] = []
        self._reindex_every = 50
        self._reindex_counter = 0

    def add(self, item: TExchangeObject):
        assert item.id not in self._items
        self._items[item.id] = item
        if item.is_open:
            self._open_items.append(item)

    def get(self, id: str) -> Optional[TExchangeObject]:
        return self._items.get(id)

    def get_open(self) -> Generator[TExchangeObject, None, None]:
        self._reindex_counter += 1
        new_open_items: Optional[List[TExchangeObject]] = None
        if self._reindex_counter % self._reindex_every == 0:
            new_open_items = []

        for item in self._open_items:
            if item.is_open:
                yield item
                if new_open_items is not None and item.is_open:
                    new_open_items.append(item)

        if new_open_items is not None:
            self._open_items = new_open_items

    def get_all(self) -> Iterable[TExchangeObject]:
        return self._items.values()



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from typing import Optional
from urllib.parse import urlencode
import datetime
import hashlib
import hmac

from basana.core import dt, pair
from basana.core.enums import OrderOperation


def get_signature(api_secret: str, qs_params: dict = {}, data: dict = {}) -> str:
    total_params = ""
    if qs_params:
        total_params += urlencode(qs_params)
    if data:
        total_params += urlencode(data)
    return hmac.new(api_secret.encode(), msg=total_params.encode("utf-8"), digestmod=hashlib.sha256).hexdigest()


def pair_to_order_book_symbol(pair: pair.Pair) -> str:
    return "{}{}".format(pair.base_symbol.upper(), pair.quote_symbol.upper())


def order_operation_to_side(operation: OrderOperation) -> str:
    return {
        OrderOperation.BUY: "BUY",
        OrderOperation.SELL: "SELL",
    }[operation]


def side_to_order_operation(side: str) -> OrderOperation:
    return {
        "BUY": OrderOperation.BUY,
        "SELL": OrderOperation.SELL,
    }[side]


def order_status_is_open(order_status: str) -> bool:
    # Not doing `order_status in ["NEW", "PARTIALLY_FILLED", "PENDING_CANCEL"]` to detect unkown states.
    is_open = {
        "NEW": True,
        "PARTIALLY_FILLED": True,
        "FILLED": False,
        "CANCELED": False,
        "PENDING_CANCEL": True,
        "REJECTED": False,
        "EXPIRED": False,
    }.get(order_status)
    assert is_open is not None, "No mapping for {} order status".format(order_status)
    return is_open


def oco_order_status_is_open(list_order_status: str) -> bool:
    # Not doing `list_order_status in ["EXECUTING"]` to detect unkown states.
    is_open = {
        "EXECUTING": True,
        "ALL_DONE": False,
        "REJECT": False,
    }.get(list_order_status)
    assert is_open is not None, "No mapping for {} list order status".format(list_order_status)
    return is_open


def get_optional_decimal(mapping: dict, key: str, skip_zero: bool) -> Optional[Decimal]:
    ret = None
    price = mapping.get(key)
    if price is not None:
        price = Decimal(price)
        ret = None if price == Decimal(0) and skip_zero else price
    return ret


def timestamp_to_datetime(timestamp: int) -> datetime.datetime:
    return datetime.datetime.fromtimestamp(timestamp / 1e3, tz=datetime.timezone.utc)


def datetime_to_timestamp(datetime: datetime.datetime) -> int:
    return int(dt.to_utc_timestamp(datetime) * 1e3)



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
import datetime

from . import helpers
from basana.core import dt, event, websockets as core_ws
from basana.core.enums import OrderOperation
from basana.core.pair import Pair


class Order:
    def __init__(self, pair: Pair, json: dict):
        #: The trading pair.
        self.pair: Pair = pair
        #: The JSON representation.
        self.json: dict = json

    @property
    def id(self) -> str:
        """The order id."""
        return str(self.json["id"])

    @property
    def datetime(self) -> datetime.datetime:
        timestamp = int(self.json["microtimestamp"]) / 1e6
        return datetime.datetime.utcfromtimestamp(timestamp).replace(tzinfo=datetime.timezone.utc)

    @property
    def amount(self) -> Decimal:
        """The order amount."""
        return Decimal(self.json["amount_at_create"])

    @property
    def amount_filled(self) -> Decimal:
        """The amount filled."""
        return self.amount - Decimal(self.json["amount_str"])

    @property
    def price(self) -> Decimal:
        """The order price."""
        return Decimal(self.json["price_str"])

    @property
    def type(self) -> OrderOperation:
        # TODO: Deprecate this property.
        return self.operation

    @property
    def operation(self) -> OrderOperation:
        """The operation."""
        return helpers.order_type_to_order_operation(int(self.json["order_type"]))


class OrderEvent(event.Event):
    """An event for order updates.

    :param when: The datetime when the event occurred. It must have timezone information set.
    :param type: The type of event. One of order_created, order_changed or order_deleted.
    :param order: The order.
    """

    def __init__(self, when: datetime.datetime, type: str, order: Order):
        super().__init__(when)
        #: The event type. One of order_created, order_changed or order_deleted.
        self.type: str = type
        #: The order.
        self.order: Order = order


# Generate Order events from websocket messages.
class WebSocketEventSource(core_ws.ChannelEventSource):
    def __init__(self, pair: Pair, producer: event.Producer):
        super().__init__(producer=producer)
        self._pair = pair

    async def push_from_message(self, message: dict):
        self.push(OrderEvent(dt.utc_now(), message["event"], Order(self._pair, message["data"])))


def get_public_channel(pair: Pair) -> str:
    return "live_orders_{}".format(helpers.pair_to_currency_pair(pair))


def get_private_channel(pair: Pair) -> str:
    return "private-my_orders_{}".format(helpers.pair_to_currency_pair(pair))



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from decimal import Decimal
from typing import Dict, Optional
import dataclasses
import datetime
import uuid

from basana.backtesting import account_balances, errors, loan_mgr, prices
from basana.backtesting.lending import base
from basana.backtesting.value_map import ValueMap, ValueMapDict


@dataclasses.dataclass
class MarginLoanConditions:
    #: The symbol for the interest.
    interest_symbol: str
    #: The interest percentage.
    interest_percentage: Decimal
    #: The interest period.
    interest_period: datetime.timedelta
    #: The minimum interest to charge.
    min_interest: Decimal
    # Minimum threshold for the value of the collateral relative to the loan amount + outstanding interests.
    margin_requirement: Decimal


class MarginLoan(base.Loan):
    def __init__(
            self, id: str, borrowed_symbol: str,  borrowed_amount: Decimal, created_at: datetime.datetime,
            conditions: MarginLoanConditions
    ):
        super().__init__(id, borrowed_symbol, borrowed_amount, created_at)
        self._conditions = conditions

    def calculate_interest(self, at: datetime.datetime, prices: prices.Prices) -> Dict[str, Decimal]:
        assert at >= self._created_at

        interest = self._conditions.interest_percentage / Decimal(100) * self.borrowed_amount
        if self._conditions.interest_period:
            time_ellapsed = at - self._created_at
            interest *= Decimal(time_ellapsed.total_seconds() / self._conditions.interest_period.total_seconds())

        # Currency conversion if interest symbol is different from borrowed symbol.
        if self._conditions.interest_symbol != self.borrowed_symbol:
            interest = prices.convert(interest, self._borrowed_symbol, self._conditions.interest_symbol)

        interest = max(interest, self._conditions.min_interest)
        return {self._conditions.interest_symbol: interest}

    def calculate_collateral(self, prices: prices.Prices) -> Dict[str, Decimal]:
        # Collateral will be managed through CheckMarginLevel at the account level.
        return {}


class MarginLoans(base.LendingStrategy):
    """
    This strategy will use the accounts assets as collateral for the loans.

    :param quote_symbol: The symbol to use to normalize balances.
    :param default_conditions: The default margin loan conditions.
    """
    def __init__(self, quote_symbol: str, default_conditions: Optional[MarginLoanConditions] = None):
        self._quote_symbol = quote_symbol
        self._conditions: Dict[str, MarginLoanConditions] = {}
        self._default_conditions = default_conditions
        self._loan_mgr: Optional[loan_mgr.LoanManager] = None
        self._exchange_ctx: Optional[base.ExchangeContext] = None

    def set_conditions(self, symbol: str, conditions: MarginLoanConditions):
        """
        Set the lending conditions for a given symbol.

        :param symbol: The symbol whose conditions are being set.
        :param conditions: The lending conditions.
        """
        self._conditions[symbol] = conditions

    def get_conditions(self, symbol: str) -> MarginLoanConditions:
        """
        Returns the lending conditions for a given symbol.

        :param symbol: The symbol.
        """
        conditions = self._conditions.get(symbol, self._default_conditions)
        if not conditions:
            raise errors.Error(f"No lending conditions for {symbol}")
        return conditions

    def set_exchange_context(self, loan_mgr: loan_mgr.LoanManager, exchange_context: base.ExchangeContext):
        self._loan_mgr = loan_mgr
        self._exchange_ctx = exchange_context
        self._exchange_ctx.account_balances.push_update_rule(CheckMarginLevel(self))

    def create_loan(self, symbol: str, amount: Decimal, created_at: datetime.datetime) -> base.Loan:
        conditions = self.get_conditions(symbol)
        return MarginLoan(uuid.uuid4().hex, symbol, amount, created_at, conditions)

    @property
    def margin_level(self) -> Decimal:
        """
        The current margin level.
        """
        assert self._exchange_ctx, "Not yet connected with the exchange"
        acc_balances = self._exchange_ctx.account_balances
        return self._calculate_margin_level(
            acc_balances.balances, acc_balances.holds, acc_balances.borrowed
        )

    def _calculate_margin_level(
            self, updated_balances: ValueMapDict, updated_holds: ValueMapDict, updated_borrowed: ValueMapDict
    ) -> Decimal:
        assert self._exchange_ctx and self._loan_mgr, "Not yet connected with the exchange"

        # Calculate used margin.
        margin_requirements = ValueMap(
            {symbol: self.get_conditions(symbol).margin_requirement for symbol in updated_borrowed}
        )
        used_margin_by_symbol = margin_requirements * updated_borrowed
        used_margin = self._exchange_ctx.prices.convert_value_map(used_margin_by_symbol, self._quote_symbol)
        if used_margin == Decimal(0):
            return Decimal(0)

        # Calculate outstanding interest.
        interest_by_symbol = ValueMap()
        for loan in self._loan_mgr.get_loans(is_open=True):
            interest_by_symbol += loan.outstanding_interest
        interest = self._exchange_ctx.prices.convert_value_map(interest_by_symbol, self._quote_symbol)

        # Calculate equity.
        equity = Decimal(0)
        for symbol, balance in updated_balances.items():
            borrowed = updated_borrowed.get(symbol, Decimal(0))
            net = balance - borrowed
            if net <= Decimal(0):
                continue

            if symbol != self._quote_symbol:
                net = self._exchange_ctx.prices.convert(net, symbol, self._quote_symbol)

            equity += net

        return equity / (used_margin + interest) * Decimal(100)

    def _check_margin_level(
            self, updated_balances: ValueMapDict, updated_holds: ValueMapDict, updated_borrowed: ValueMapDict
    ):
        margin_level = self._calculate_margin_level(updated_balances, updated_holds, updated_borrowed)
        if margin_level > Decimal(0) and margin_level < Decimal(100):
            raise errors.NotEnoughBalance(f"Margin level too low {margin_level}")


class CheckMarginLevel(account_balances.UpdateRule):
    def __init__(self, margin_loans: MarginLoans):
        self._margin_loans = margin_loans

    def check(self, updated_balances: ValueMapDict, updated_holds: ValueMapDict, updated_borrowed: ValueMapDict):
        self._margin_loans._check_margin_level(updated_balances, updated_holds, updated_borrowed)



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional
import argparse
import datetime
import sys

import aiohttp
import asyncio

from basana.core import dt, token_bucket
from basana.external.binance import client, helpers


period_to_step = {
    "1s": 1,
    "1m": 60,
    "3m": 3 * 60,
    "5m": 5 * 60,
    "15m": 15 * 60,
    "30m": 30 * 60,
    "1h": 3600,
    "2h": 2 * 3600,
    "4h": 4 * 3600,
    "6h": 6 * 3600,
    "8h": 8 * 3600,
    "12h": 12 * 3600,
    "1d": 86400,
    "3d": 3 * 86400,
    "1w": 7 * 86400,
    "1M": 31 * 86400,
}


def parse_date(date: str):
    return datetime.datetime.combine(
        datetime.date.fromisoformat(date), datetime.time()
    ).replace(tzinfo=datetime.timezone.utc)


class Candlestick:
    def __init__(self, candlestick: list):
        self.open_timestamp = candlestick[0]
        self.open = candlestick[1]
        self.high = candlestick[2]
        self.low = candlestick[3]
        self.close = candlestick[4]
        self.volume = candlestick[5]
        self.close_timestamp = candlestick[6]


def to_binance_currency_pair(currency_pair: str):
    parts = currency_pair.upper().split("/")
    return "".join(parts)


class CSVWriter:
    def __init__(self, output_file: Optional[str]):
        self._header_written = False
        self._output_file = open(output_file, "w") if output_file else sys.stdout

    def write_candlestick(self, candlestick: Candlestick):
        if not self._header_written:
            print("datetime,open,high,low,close,volume", file=self._output_file)
            self._header_written = True

        dt_col = datetime.datetime.fromtimestamp(candlestick.open_timestamp / 1000, tz=datetime.timezone.utc)
        print(",".join([
            str(dt_col.replace(tzinfo=None)),
            candlestick.open, candlestick.high, candlestick.low, candlestick.close, candlestick.volume
        ]), file=self._output_file)


async def main(params: Optional[List[str]] = None, config_overrides: dict = {}):
    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--currency-pair", help="The currency pair.", required=True)
    parser.add_argument(
        "-p", "--period", help="The period for the bars.", choices=period_to_step.keys(), required=True
    )
    parser.add_argument(
        "-s", "--start", help="The starting date YYYY-MM-DD format. Included in the range.", required=True
    )
    parser.add_argument(
        "-e", "--end", help="The ending date YYYY-MM-DD format. Included in the range.", required=True
    )
    parser.add_argument("-o", "--output", help="The output file.", required=False, default=None)
    args = parser.parse_args(args=params)

    step = period_to_step[args.period]
    start = parse_date(args.start)
    end = parse_date(args.end)
    assert start <= end, "Invalid start/end"
    # start/end are set as dates so to get past the end we need to use a step >= 1 day.
    past_the_end = end + datetime.timedelta(seconds=max(period_to_step["1d"], step))

    start_ts = helpers.datetime_to_timestamp(start)
    past_the_end_ts = helpers.datetime_to_timestamp(past_the_end)
    step = step * 1000

    tb = token_bucket.TokenBucketLimiter(10, 1)
    now_ts = helpers.datetime_to_timestamp(dt.utc_now())
    writer = CSVWriter(args.output)
    async with aiohttp.ClientSession() as session:
        cli = client.APIClient(session=session, tb=tb, config_overrides=config_overrides)
        eof = False
        currency_pair = to_binance_currency_pair(args.currency_pair)
        while not eof:
            response = await cli.get_candlestick_data(
                currency_pair, args.period, start_time=start_ts, end_time=past_the_end_ts, limit=1000
            )
            eof = True
            for candlestick in response:
                eof = False
                candlestick = Candlestick(candlestick)
                start_ts = max(start_ts, candlestick.open_timestamp)
                if candlestick.open_timestamp >= past_the_end_ts or candlestick.close_timestamp >= now_ts:
                    continue
                writer.write_candlestick(candlestick)
            if not eof:
                start_ts += step
                eof = start_ts >= past_the_end_ts


if __name__ == "__main__":  # pragma: no cover
    asyncio.run(main())



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import json
import logging

from . import dt


@contextlib.contextmanager
def backtesting_log_mode(dispatcher):
    old_factory = logging.getLogRecordFactory()

    def record_factory(*args, **kwargs):
        record_dt = dispatcher.now()
        record = old_factory(*args, **kwargs)
        record.created = dt.to_utc_timestamp(record_dt)
        record.msecs = int(record_dt.microsecond / 1000)
        return record

    logging.setLogRecordFactory(record_factory)
    yield
    logging.setLogRecordFactory(old_factory)


# https://docs.python.org/3/howto/logging-cookbook.html#implementing-structured-logging
class StructuredMessage:
    def __init__(self, message, /, **kwargs):
        self.message = message
        self.kwargs = kwargs

    def __str__(self):
        return "{} {}".format(self.message, json.dumps(self.kwargs, default=str))



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import cast, Any, Awaitable, Callable, List, Optional
import asyncio
import json
import logging

import aiohttp

from basana.core import logs, websockets as core_ws
from basana.core.config import get_config_value
from basana.external.bitstamp import client, config


logger = logging.getLogger(__name__)

CHANNEL_EVENTS = {"trade", "data", "order_created", "order_changed", "order_deleted"}


class WebSocketClient(core_ws.WebSocketClient):
    def __init__(self, session: Optional[aiohttp.ClientSession] = None, config_overrides: dict = {}):
        super().__init__(
            get_config_value(config.DEFAULTS, "api.websockets.base_url", overrides=config_overrides),
            session=session, config_overrides=config_overrides,
            heartbeat=get_config_value(config.DEFAULTS, "api.websockets.heartbeat", overrides=config_overrides)
        )

    async def handle_message(self, message: dict) -> bool:
        coro = None
        event = message.get("event", "")

        # Is it an admin, non-channel, message ?
        message_handler = cast(
            Optional[Callable[[dict], Awaitable[Any]]],
            {
                "bts:request_reconnect": self._on_bts_request_reconnect,
                "bts:subscription_succeeded": self._on_bts_subscription_succeeded,
                "bts:error": self.on_error,
                "bts:subscription_failed": self.on_error,
            }.get(event)
        )
        if message_handler:
            coro = message_handler(message)
        # Is it a channel message ?
        elif event in CHANNEL_EVENTS \
                and (channel := message.get("channel")) \
                and (event_source := self.get_channel_event_source(channel)):
            coro = event_source.push_from_message(message)

        ret = False
        if coro:
            await coro
            ret = True
        return ret

    async def _on_bts_request_reconnect(self, message: dict):
        logger.debug(logs.StructuredMessage("Reconnection requested", message=message))
        self.schedule_reconnection()

    async def _on_bts_subscription_succeeded(self, message: dict):
        logger.debug(logs.StructuredMessage("Subscription suceeded", message=message))


class PublicWebSocketClient(WebSocketClient):
    def __init__(self, session: Optional[aiohttp.ClientSession] = None, config_overrides: dict = {}):
        super().__init__(session=session, config_overrides=config_overrides)

    async def subscribe_to_channels(self, channels: List[str], ws_cli: aiohttp.ClientWebSocketResponse):
        logger.debug(logs.StructuredMessage("Subscribing", src=self, channels=channels))
        await asyncio.gather(*[
            ws_cli.send_str(json.dumps({
                "event": "bts:subscribe",
                "data": {
                    "channel": channel
                }
            }))
            for channel in channels
        ])


class PrivateWebSocketClient(WebSocketClient):
    def __init__(
            self, api_key: str, api_secret: str, session: Optional[aiohttp.ClientSession] = None,
            config_overrides: dict = {}
    ):
        super().__init__(session=session, config_overrides=config_overrides)
        self._client = client.APIClient(api_key, api_secret, session=session, config_overrides=config_overrides)

    async def subscribe_to_channels(self, channels: List[str], ws_cli: aiohttp.ClientWebSocketResponse):
        logger.debug(logs.StructuredMessage("Subscribing", src=self, channels=channels))
        logger.debug(logs.StructuredMessage("Authenticating", src=self))
        websockets_token = await self._client.get_websocket_auth_token()

        await asyncio.gather(*[
            ws_cli.send_str(json.dumps({
                "event": "bts:subscribe",
                "data": {
                    "auth": websockets_token["token"],
                    "channel": "{}-{}".format(channel, websockets_token["user_id"])
                }
            }))
            for channel in channels
        ])



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import time


class TokenBucketLimiter:
    """This class implements a token bucket algorithm, useful for throttling requests.

    :param tokens_per_period: The maximum amount of tokens per perdiod.
    :param period_duration: The period duration in seconds.
    :param initial_tokens: The initial amount of tokens.
    """

    def __init__(self, tokens_per_period: float, period_duration: int, initial_tokens=0):
        assert tokens_per_period > 0
        assert period_duration > 0
        assert initial_tokens >= 0

        self._tokens_per_period = tokens_per_period
        self._period_duration = period_duration
        self._tokens = initial_tokens
        self._last = time.time()

    @property
    def tokens(self) -> int:
        return max(int(self._tokens), 0)

    @property
    def tokens_per_period(self) -> float:
        return self._tokens_per_period

    @property
    def period_duration(self) -> int:
        return self._period_duration

    def consume(self) -> float:
        """Consumes one token and returns the time to wait before using it."""

        # Refill pool of tokens.
        now = time.time()
        lapse = now - self._last
        self._last = now
        self._tokens += lapse / self._period_duration * self._tokens_per_period
        if self._tokens > self._tokens_per_period:
            self._tokens = self._tokens_per_period

        # Consume one token.
        self._tokens -= 1

        if self._tokens >= 0:
            return 0.0
        else:
            return -self._tokens / self._tokens_per_period * self._period_duration

    async def wait(self):
        await asyncio.sleep(self.consume())



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from typing import Dict, Optional, Sequence
import collections
import datetime

from . import helpers
from basana.core.enums import OrderOperation


class Balance:
    def __init__(self, json: dict):
        self.json = json

    @property
    def available(self) -> Decimal:
        """The available balance."""
        return Decimal(self.json["free"])

    @property
    def total(self) -> Decimal:
        """The total balance (available + locked)."""
        return self.available + self.locked

    @property
    def locked(self) -> Decimal:
        """The locked balance."""
        return Decimal(self.json["locked"])


class Trade:
    def __init__(self, json: dict):
        self.json = json

    @property
    def id(self) -> str:
        """The trade id."""
        return str(self.json["id"])

    @property
    def order_id(self) -> str:
        """The order id."""
        return str(self.json["orderId"])

    @property
    def datetime(self) -> datetime.datetime:
        return helpers.timestamp_to_datetime(self.json["time"])

    @property
    def is_best_match(self) -> bool:
        return self.json["isBestMatch"]

    @property
    def is_buyer(self) -> bool:
        return self.json["isBuyer"]

    @property
    def is_maker(self) -> bool:
        return self.json["isMaker"]

    @property
    def price(self) -> Decimal:
        return Decimal(self.json["price"])

    @property
    def amount(self) -> Decimal:
        return Decimal(self.json["qty"])

    @property
    def quote_amount(self) -> Decimal:
        return Decimal(self.json["quoteQty"])

    @property
    def commission(self) -> Decimal:
        return Decimal(self.json["commission"])

    @property
    def commission_asset(self) -> str:
        return self.json["commissionAsset"]


class OrderWrapper:
    def __init__(self, json: dict):
        self.json = json

    @property
    def id(self) -> str:
        """The order id."""
        return str(self.json["orderId"])

    @property
    def client_order_id(self) -> str:
        """The client order id."""
        return self.json["clientOrderId"]

    @property
    def order_list_id(self) -> Optional[str]:
        """The order list id."""
        ret = self.json.get("orderListId")
        ret = None if ret in [None, -1] else str(ret)
        return ret

    @property
    def status(self) -> str:
        """The status.

        Check **Order status** in
        https://developers.binance.com/docs/binance-spot-api-docs/enums#order-status-status.
        """
        return self.json["status"]

    @property
    def is_open(self) -> bool:
        """True if the order is open, False otherwise."""
        return helpers.order_status_is_open(self.status)

    @property
    def amount(self) -> Decimal:
        """The amount."""
        return Decimal(self.json["origQty"])

    @property
    def amount_filled(self) -> Decimal:
        """The amount filled."""
        return Decimal(self.json["executedQty"])

    @property
    def quote_amount_filled(self) -> Decimal:
        """The amount filled in quote units."""
        return Decimal(self.json["cummulativeQuoteQty"])

    @property
    def limit_price(self) -> Optional[Decimal]:
        """The limit price."""
        return helpers.get_optional_decimal(self.json, "price", True)

    @property
    def stop_price(self) -> Optional[Decimal]:
        """The stop price."""
        return helpers.get_optional_decimal(self.json, "stopPrice", True)

    @property
    def time_in_force(self) -> Optional[str]:
        """The time in force.

        Check **Time in force** in
        https://developers.binance.com/docs/binance-spot-api-docs/enums#time-in-force-timeinforce.
        """
        return self.json.get("timeInForce")


class OrderInfo(OrderWrapper):
    def __init__(self, json: dict, trades: Sequence[Trade]):
        super().__init__(json)
        self.trades = trades
        self._fees: Dict[str, Decimal] = collections.defaultdict(Decimal)
        for trade in trades:
            if trade.commission:
                self._fees[trade.commission_asset] += trade.commission

    @property
    def operation(self) -> OrderOperation:
        """The operation."""
        return helpers.side_to_order_operation(self.json["side"])

    @property
    def amount_remaining(self) -> Decimal:
        """The amount remaining to be filled."""
        return self.amount - self.amount_filled

    @property
    def fill_price(self) -> Optional[Decimal]:
        """The fill price."""
        ret = None
        if self.amount_filled:
            ret = self.quote_amount_filled / self.amount_filled
        return ret

    @property
    def fees(self) -> Dict[str, Decimal]:
        """The fees."""
        return self._fees


class Fill:
    def __init__(self, json: dict):
        self.json = json

    @property
    def price(self) -> Decimal:
        """The price."""
        return Decimal(self.json["price"])

    @property
    def amount(self) -> Decimal:
        """The amount."""
        return Decimal(self.json["qty"])

    @property
    def commission(self) -> Decimal:
        """The commission."""
        return Decimal(self.json["commission"])

    @property
    def commission_asset(self) -> str:
        """The commission asset."""
        return self.json["commissionAsset"]


class CreatedOrder:
    def __init__(self, json: dict):
        self.json = json

    @property
    def id(self) -> str:
        """The order id."""
        return str(self.json["orderId"])

    @property
    def datetime(self) -> datetime.datetime:
        """The creation datetime."""
        return helpers.timestamp_to_datetime(self.json["transactTime"])

    @property
    def client_order_id(self) -> str:
        """The client order id."""
        return self.json["clientOrderId"]

    @property
    def limit_price(self) -> Optional[Decimal]:
        """The limit price.

        Only available for RESULT / FULL responses.
        """
        return helpers.get_optional_decimal(self.json, "price", True)

    @property
    def amount(self) -> Optional[Decimal]:
        """The amount.

        Only available for RESULT / FULL responses.
        """
        return helpers.get_optional_decimal(self.json, "origQty", False)

    @property
    def amount_filled(self) -> Optional[Decimal]:
        """The amount filled.

        Only available for RESULT / FULL responses.
        """
        return helpers.get_optional_decimal(self.json, "executedQty", False)

    @property
    def quote_amount_filled(self) -> Optional[Decimal]:
        """The amount filled in quote units.

        Only available for RESULT / FULL responses.
        """
        return helpers.get_optional_decimal(self.json, "cummulativeQuoteQty", False)

    @property
    def status(self) -> Optional[str]:
        """The status.

        Only available for RESULT / FULL responses.
        Check **Order status** in https://developers.binance.com/docs/binance-spot-api-docs/enums#order-status-status.
        """
        return self.json.get("status")

    @property
    def time_in_force(self) -> Optional[str]:
        """The time in force.

        Only available for RESULT / FULL responses.
        Check **Time in force** in
        https://developers.binance.com/docs/binance-spot-api-docs/enums#time-in-force-timeinforce.
        """
        return self.json.get("timeInForce")

    @property
    def is_open(self) -> bool:
        """True if the order is open, False otherwise.

        Only available for RESULT / FULL responses.
        """
        assert self.status is not None, "status not set"
        return helpers.order_status_is_open(self.status)


class CanceledOrder(OrderWrapper):
    @property
    def operation(self) -> OrderOperation:
        """The operation."""
        return helpers.side_to_order_operation(self.json["side"])

    @property
    def type(self) -> str:
        """The type of order.

        Check **Order types** in
        https://developers.binance.com/docs/binance-spot-api-docs/enums#order-types-ordertypes-type.
        """
        return self.json["type"]


class OpenOrder(OrderWrapper):
    @property
    def datetime(self) -> datetime.datetime:
        """The creation datetime."""
        return helpers.timestamp_to_datetime(self.json["time"])

    @property
    def operation(self) -> OrderOperation:
        """The operation."""
        return helpers.side_to_order_operation(self.json["side"])

    @property
    def type(self) -> str:
        """The type of order.

        Check **Order types** in
        https://developers.binance.com/docs/binance-spot-api-docs/enums#order-types-ordertypes-type.
        """
        return self.json["type"]


class OCOOrderWrapper:
    def __init__(self, json: dict):
        self.json = json

    @property
    def order_list_id(self) -> str:
        """The order list id."""
        return str(self.json["orderListId"])

    @property
    def client_order_list_id(self) -> str:
        """A client id for the order list."""
        return str(self.json["listClientOrderId"])

    @property
    def datetime(self) -> datetime.datetime:
        """The creation datetime."""
        return helpers.timestamp_to_datetime(self.json["transactionTime"])

    @property
    def is_open(self) -> bool:
        """True if the order is open, False otherwise."""
        return helpers.oco_order_status_is_open(self.json["listOrderStatus"])

    @property
    def limit_order_id(self) -> str:
        """The id for the limit order."""
        order_ids = [
            str(order["orderId"])
            for order in self.json.get("orderReports", [])
            if order["type"] in ("LIMIT", "LIMIT_MAKER")
        ]
        return order_ids[0]

    @property
    def stop_loss_order_id(self) -> str:
        """The id for the stop loss order."""
        order_ids = [
            str(order["orderId"])
            for order in self.json.get("orderReports", [])
            if order["type"] in ("STOP_LOSS", "STOP_LOSS_LIMIT")
        ]
        return order_ids[0]


class CreatedOCOOrder(OCOOrderWrapper):
    pass


class OCOOrderInfo(OCOOrderWrapper):
    pass


class CanceledOCOOrder(OCOOrderWrapper):
    pass



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from typing import List
import abc
import itertools

from basana.backtesting import errors
from basana.backtesting.value_map import ValueMap, ValueMapDict


class UpdateRule(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def check(self, updated_balances: ValueMap, updated_holds: ValueMap, updated_borrowed: ValueMap):
        raise NotImplementedError()


class NonZero(UpdateRule):
    def check(self, updated_balances: ValueMap, updated_holds: ValueMap, updated_borrowed: ValueMap):
        # balance >= 0
        for symbol, value in updated_balances.items():
            if value < Decimal(0):
                raise errors.NotEnoughBalance(f"Not enough {symbol} available")
        # hold >= 0
        for symbol, value in updated_holds.items():
            if value < Decimal(0):
                raise errors.Error(f"hold update amount for {symbol} is invalid")
        # borrowed >= 0
        for symbol, value in updated_borrowed.items():
            if value < Decimal(0):
                raise errors.Error(f"borrowed update amount for {symbol} is invalid")


class ValidHold(UpdateRule):
    # * hold <= balance
    def check(self, updated_balances: ValueMap, updated_holds: ValueMap, updated_borrowed: ValueMap):
        symbols = set(itertools.chain(updated_holds.keys(), updated_balances.keys()))
        for symbol in symbols:
            updated_hold = updated_holds.get(symbol, Decimal(0))
            updated_balance = updated_balances.get(symbol, Decimal(0))
            if updated_hold > updated_balance:
                raise errors.NotEnoughBalance(f"Not enough {symbol} available to hold")


class AccountBalances:
    def __init__(self, initial_balances: ValueMapDict):
        self.balances = ValueMap({
            symbol: balance for symbol, balance in initial_balances.items() if balance >= 0
        })
        self.holds = ValueMap()
        self.borrowed = ValueMap({
            symbol: -balance for symbol, balance in initial_balances.items() if balance < 0
        })
        self._update_rules: List[UpdateRule] = [
            NonZero(),
            ValidHold()
        ]

    def push_update_rule(self, update_rule: UpdateRule):
        self._update_rules.append(update_rule)

    def update(
            self, balance_updates: ValueMapDict = {}, hold_updates: ValueMapDict = {},
            borrowed_updates: ValueMapDict = {}
    ):
        updated_balances = self.balances + balance_updates
        updated_holds = self.holds + hold_updates
        updated_borrowed = self.borrowed + borrowed_updates

        for rule in self._update_rules:
            rule.check(updated_balances, updated_holds, updated_borrowed)

        # Update if no error ocurred.
        self.balances = updated_balances
        self.holds = updated_holds
        self.borrowed = updated_borrowed

    def get_symbols(self) -> List[str]:
        symbols = set(self.balances.keys())
        symbols.update(self.holds.keys())
        symbols.update(self.borrowed.keys())
        return list(symbols)

    def get_available_balance(self, symbol: str) -> Decimal:
        return self.balances.get(symbol, Decimal(0)) - self.holds.get(symbol, Decimal(0))

    def get_balance_on_hold(self, symbol: str) -> Decimal:
        return self.holds.get(symbol, Decimal(0))

    def get_borrowed_balance(self, symbol: str) -> Decimal:
        return self.borrowed.get(symbol, Decimal(0))



# Basana
#
# Copyright 2022 Gabriel Martin Becedillas Ruiz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from decimal import Decimal
from typing import Any, List, Optional
import asyncio
import datetime
import logging

import aiohttp

from . import client, helpers
from basana.core import dt, event, logs, token_bucket, websockets as core_ws
from basana.core.pair import Pair


logger = logging.getLogger(__name__)


@dataclass
class Entry:
    #: The price.
    price: Decimal
    #: The volume.
    volume: Decimal


# https://www.bitstamp.net/api/#order-book
class OrderBook:
    def __init__(self, pair: Pair, json: dict):
        #: The trading pair.
        self.pair: Pair = pair
        #: The JSON representation.
        self.json: dict = json

    @property
    def datetime(self) -> datetime.datetime:
        timestamp = int(self.json["microtimestamp"]) / 1e6
        return datetime.datetime.fromtimestamp(timestamp, tz=datetime.timezone.utc)

    @property
    def bids(self) -> List[Entry]:
        """Returns the top bid entries."""
        return [
            Entry(price=Decimal(entry[0]), volume=Decimal(entry[1])) for entry in self.json["bids"]
        ]

    @property
    def asks(self) -> List[Entry]:
        """Returns the top ask entries."""
        return [
            Entry(price=Decimal(entry[0]), volume=Decimal(entry[1])) for entry in self.json["asks"]
        ]


class OrderBookEvent(event.Event):
    """An event for order book updates.

    :param when: The datetime when the event occurred. It must have timezone information set.
    :param order_book: The updated order book.
    """

    def __init__(self, when: datetime.datetime, order_book: OrderBook):
        super().__init__(when)
        #: The order book.
        self.order_book: OrderBook = order_book


class PollOrderBook(event.FifoQueueEventSource, event.Producer):
    def __init__(
            self, pair: Pair, interval: float, group: Optional[int] = None,
            session: Optional[aiohttp.ClientSession] = None, tb: Optional[token_bucket.TokenBucketLimiter] = None,
            config_overrides: dict = {}
    ):
        assert interval > 0, "Invalid interval"

        super().__init__(producer=self)
        self.pair = pair
        self._interval = interval
        self._group = group
        self._client = client.APIClient(session=session, tb=tb, config_overrides=config_overrides)

    async def _fetch_and_push(self, currency_pair: str):
        order_book_json = await self._client.get_order_book(currency_pair, group=self._group)
        self.push(OrderBookEvent(dt.utc_now(), OrderBook(self.pair, order_book_json)))

    async def main(self):
        currency_pair = helpers.pair_to_currency_pair(self.pair)
        while True:
            try:
                await self._fetch_and_push(currency_pair)
            except Exception as e:
                await self.on_error(e)
            await asyncio.sleep(self._interval)

    async def on_error(self, error: Any):
        logger.error(logs.StructuredMessage("Error polling order book", channel=self.pair, error=error))


# Generate OrderBookEvent events from websocket messages.
class WebSocketEventSource(core_ws.ChannelEventSource):
    def __init__(self, pair: Pair, producer: event.Producer):
        super().__init__(producer=producer)
        self._pair = pair

    async def push_from_message(self, message: dict):
        self.push(OrderBookEvent(dt.utc_now(), OrderBook(self._pair, message["data"])))


def get_channel(pair: Pair) -> str:
    return "order_book_{}".format(helpers.pair_to_currency_pair(pair))
